from pyvcam import pvc
from pyvcam import constants as const

from copy import deepcopy
import os
import time
import numpy as np


class Camera:
    """Models a class currently connected to the system.

    Attributes:
        __name(str): String containing the name of the camera.
        __handle(int): The current camera's handle.
        __is_open(bool): True if camera is opened, False otherwise.

        __exposure_bytes(int): How large the buffer for live imaging needs to be.

        __mode(int): The bit-wise or between exposure mode and expose out mode.
        __exp_time(int): Integer representing the exposure time to be used for captures.
    """

    class ReversibleEnumDict(dict):
        # Helper class to populate enumerated parameter dictionaries and ease conversion of keys and values.
        # The param_id must support enum attribute
        # This dictionary will accept both keys and values to the __getitem__ operator, rather than just keys like
        # regular dictionaries. If a value is provided, the matching key is returned.
        def __init__(self, name, camera_instance, param_id):
            try:
                enumDict = camera_instance.read_enum(param_id)
            except AttributeError:
                enumDict = {}

            super(Camera.ReversibleEnumDict, self).__init__(enumDict)
            self.name = name

        def __getitem__(self, keyOrValue):
            try:
                if isinstance(keyOrValue, str):
                    return super(Camera.ReversibleEnumDict, self).__getitem__(keyOrValue)
                else:
                    return [key for key, item_value in self.items() if keyOrValue == item_value][0]
            except KeyError:
                raise ValueError('Invalid key: {0} for {1} - Available keys are: {2}'.format(keyOrValue, self.name, list(self.keys())))
            except IndexError:
                raise ValueError('Invalid value: {0} for {1} - Available values are: {2}'.format(keyOrValue, self.name, list(self.values())))

    class RegionOfInterest:

        def __init__(self, s1, s2, sbin, p1, p2, pbin):
            self.s1 = s1
            self.s2 = s2 - ((s2 - s1 + 1) % sbin)  # Clip partially binned pixels
            self.sbin = sbin
            self.p1 = p1
            self.p2 = p2 - ((p2 - p1 + 1) % pbin)  # Clip partially binned pixels
            self.pbin = pbin

        def __eq__(self, other):
            if not isinstance(other, Camera.RegionOfInterest):
                raise NotImplementedError

            return self.s1 == other.s1 and self.s2 == other.s2 and self.p1 == other.p1 and self.p2 == other.p2

        def checkOverlap(self, other):
            return (self.s1 <= other.s1 <= self.s1 and self.p1 <= other.p1 <= self.p1) or \
                   (self.s2 <= other.s2 <= self.s2 and self.p2 <= other.p2 <= self.p2)

        @property
        def shape(self):
            return int((self.s2 - self.s1 + 1) / self.sbin), int((self.p2 - self.p1 + 1) / self.pbin)

    WAIT_FOREVER = -1

    def __init__(self, name):
        """NOTE: CALL Camera.detect_camera() to get a camera object."""
        self.__name = name
        self.__handle = -1
        self.__is_open = False
        self.__acquisition_mode = None

        # Memory for live circular buffer
        self.__exposure_bytes = None

        # Exposure Settings
        self.__mode = None
        self.__exp_time = 0

        # Regions of interest
        self.__defaultRoi = None  # The default RegionOfInterest object
        self.__rois = []  # A list of RegionOfInterest objects

    def __repr__(self):
        return self.__name

    @staticmethod
    def get_available_camera_names():
        """Gets the name for each available camera.

        Returns:
           List of camera names, sorted by index.
        """
        ret = []
        total = pvc.get_cam_total()

        for index in range(total):
            ret.append(pvc.get_cam_name(index))

        return ret

    @classmethod
    def detect_camera(cls):
        """Detects and creates a new Camera object.

        Returns:
            A Camera object generator.
        """
        cam_count = 0
        total = pvc.get_cam_total()
        while cam_count < total:
            try:
                yield Camera(pvc.get_cam_name(cam_count))
                cam_count += 1
            except RuntimeError:
                raise RuntimeError('Failed to create a detected camera.')

    @classmethod
    def select_camera(cls, name):
        """Select camera by name and creates a new Camera object.

        Returns:
            A Camera object.
        """
        total = pvc.get_cam_total()
        for index in range(total):
            try:
                if name == pvc.get_cam_name(index):
                    return Camera(name)
            except RuntimeError:
                raise RuntimeError('Failed to create a detected camera.')

        raise RuntimeError('Failed to create a detected camera. Invalid name')

    def open(self):
        """Opens the camera.

        Side Effect(s):
            - changes self.__handle upon successful call to pvc module.
            - changes self.__is_open to True
            - changes self.__roi to sensor's full frame

        Returns:
            None
        """

        try:
            self.__handle = pvc.open_camera(self.__name)
            self.__is_open = True
        except:
            raise RuntimeError('Failed to open camera.')

        # If the camera is frame transfer capable, then set its p-mode to
        # frame transfer, otherwise set it to normal mode.
        try:
            self.get_param(const.PARAM_FRAME_CAPABLE, const.ATTR_CURRENT)
            self.set_param(const.PARAM_PMODE, const.PMODE_FT)
        except AttributeError:
            self.set_param(const.PARAM_PMODE, const.PMODE_NORMAL)

        # Set ROI to full frame with no binning
        self.__defaultRoi = Camera.RegionOfInterest(0, self.sensor_size[0] - 1, 1, 0, self.sensor_size[1] - 1, 1)
        self.reset_rois()

        # Setup correct mode
        self.__exp_mode = self.get_param(const.PARAM_EXPOSURE_MODE)
        if self.check_param(const.PARAM_EXPOSE_OUT_MODE):
            self.__exp_out_mode = self.get_param(const.PARAM_EXPOSE_OUT_MODE)
        else:
            self.__exp_out_mode = 0

        self.__mode = self.__exp_mode | self.__exp_out_mode

        # Populate enumerated values
        self.__centroids_modes = Camera.ReversibleEnumDict('centroids_modes', self, const.PARAM_CENTROIDS_MODE)
        self.__clear_modes = Camera.ReversibleEnumDict('clear_modes', self, const.PARAM_CLEAR_MODE)
        self.__exp_modes = Camera.ReversibleEnumDict('exp_modes', self, const.PARAM_EXPOSURE_MODE)
        self.__exp_out_modes = Camera.ReversibleEnumDict('exp_out_modes', self, const.PARAM_EXPOSE_OUT_MODE)
        self.__exp_resolutions = Camera.ReversibleEnumDict('exp_resolutions', self, const.PARAM_EXP_RES)
        self.__prog_scan_modes = Camera.ReversibleEnumDict('prog_scan_modes', self, const.PARAM_SCAN_MODE)
        self.__prog_scan_dirs = Camera.ReversibleEnumDict('prog_scan_dirs', self, const.PARAM_SCAN_DIRECTION)

        # Learn ports, speeds and gains
        self.__port_speed_gain_table = {}
        for port_name, port_value in self.read_enum(const.PARAM_READOUT_PORT).items():
            self.__port_speed_gain_table[port_name] = {'port_value': port_value}
            self.readout_port = port_value
            num_speeds = self.get_param(const.PARAM_SPDTAB_INDEX, const.ATTR_COUNT)
            for speed_index in range(num_speeds):
                speed_name = 'Speed_' + str(speed_index)
                self.speed_table_index = speed_index

                gain_min = self.get_param(const.PARAM_GAIN_INDEX, const.ATTR_MIN)
                gain_max = self.get_param(const.PARAM_GAIN_INDEX, const.ATTR_MAX)
                gain_increment = self.get_param(const.PARAM_GAIN_INDEX, const.ATTR_INCREMENT)

                numGains = int((gain_max - gain_min) / gain_increment + 1)
                gains = [(gain_min + i * gain_increment) for i in range(numGains)]

                self.__port_speed_gain_table[port_name].update({speed_name: {'speed_index': speed_index, 'pixel_time': self.pix_time, 'bit_depth': self.bit_depth, 'gain_range': gains}})

                for gain_index in gains:
                    self.gain = gain_index
                    try:
                        gain_name = self.get_param(const.PARAM_GAIN_NAME, const.ATTR_CURRENT)
                    except:
                        gain_name = 'Gain_' + str(gain_index)

                    self.__port_speed_gain_table[port_name][speed_name][gain_name] = {'gain_index': gain_index}

        # Reset speed table back to default
        self.readout_port = 0
        self.speed_table_index = 0

        # Learn post processing features
        self.__post_processing_table = {}
        try:
            featureCount = self.get_param(const.PARAM_PP_INDEX, const.ATTR_COUNT)
            for featureIndex in range(featureCount):

                self.set_param(const.PARAM_PP_INDEX, featureIndex)
                featureId = self.get_param(const.PARAM_PP_FEAT_ID)
                featureName = self.get_param(const.PARAM_PP_FEAT_NAME)
                self.__post_processing_table[featureName] = {}

                paramCount = self.get_param(const.PARAM_PP_PARAM_INDEX, const.ATTR_COUNT)
                for paramIndex in range(paramCount):
                    self.set_param(const.PARAM_PP_PARAM_INDEX, paramIndex)
                    paramId = self.get_param(const.PARAM_PP_PARAM_ID)
                    paramName = self.get_param(const.PARAM_PP_PARAM_NAME)
                    paramMin = self.get_param(const.PARAM_PP_PARAM, const.ATTR_MIN)
                    paramMax = self.get_param(const.PARAM_PP_PARAM, const.ATTR_MAX)
                    self.__post_processing_table[featureName][paramName] = {'feature_index': featureIndex, 'feature_id': featureId, 'param_index': paramIndex, 'param_id': paramId, 'param_min': paramMin, 'param_max': paramMax}

        except AttributeError:
            pass

    def close(self):
        """Closes the camera.

        Side Effect(s):
            - changes self.__handle upon successful call to pvc module.
            - changes self.__is_open to False

        Returns:
            None
        """
        try:
            pvc.close_camera(self.__handle)
            self.__handle = -1
            self.__is_open = False
        except:
            raise RuntimeError('Failed to close camera.')

    def check_frame_status(self):
        """Gets the frame transfer status. Will raise an exception if called prior to initiating acquisition

        Parameter(s):
            None

        Returns:
            String representation of PL_IMAGE_STATUSES enum from pvcam.h
            'READOUT_NOT_ACTIVE' - The system is @b idle, no data is expected. If any arrives, it will be discarded.
            'EXPOSURE_IN_PROGRESS' - The data collection routines are @b active. They are waiting for data to arrive, but none has arrived yet.
            'READOUT_IN_PROGRESS' - The data collection routines are @b active. The data has started to arrive.
            'READOUT_COMPLETE' - All frames available in sequnece mode.
            'FRAME_AVAILABLE' - At least one frame is available in live mode
            'READOUT_FAILED' - Something went wrong.
        """

        status = pvc.check_frame_status(self.__handle)

        # TODO:  pvcam currently returns FRAME_AVAILABLE/READOUT_COMPLETE after a sequence is finished. Until this behavior is resolved,
        #        force status to READOUT_NOT_ACTIVE while not acquiring
        status = 'READOUT_NOT_ACTIVE' if (self.__acquisition_mode is None) else status
        return status

    def get_param(self, param_id, param_attr=const.ATTR_CURRENT):
        """Gets the current value of a specified parameter.

        Parameter(s):
            param_id (int): The parameter to get. Refer to constants.py for
                            defined constants for each parameter.
            param_attr (int): The desired attribute of the parameter to
                              identify. Refer to constants.py for defined
                              constants for each attribute.

        Returns:
            Value of specified parameter.
        """
        return pvc.get_param(self.__handle, param_id, param_attr)

    def set_param(self, param_id, value):
        """Sets a specified setting of a camera to a specified value.

        Note that pvc will raise a RuntimeError if the camera setting can not be
        applied. Pvc will also raise a ValueError if the supplied arguments are
        invalid for the specified parameter.

        Side Effect(s):
            - changes camera's internal setting.

        Parameters:
            param_id (int): An int that corresponds to a camera setting. Refer to
                            constants.py for valid parameter values.
            value (Varies): The value to set the camera setting to.
        """
        pvc.set_param(self.__handle, param_id, value)

    def check_param(self, param_id):
        """Checks if a specified setting of a camera is available to read/ modify.

        Side Effect(s):
            - None

        Parameters:
            param_id (int): An int that corresponds to a camera setting. Refer to
                            constants.py for valid parameter values.

        Returns:
            Boolean true if available, false if not
        """
        return pvc.check_param(self.__handle, param_id)

    def read_enum(self, param_id):
        """ Returns all settings names paired with their values of a parameter.

        Parameter:
            param_id (int):  The parameter ID.

        Returns:
            A dictionary containing strings mapped to values.
        """
        return pvc.read_enum(self.__handle, param_id)

    def reset_pp(self):
        """Resets the post-processing settings to default.

        Returns:
            None
        """
        try:
            pvc.reset_pp(self.__handle)
        except:
            raise RuntimeError('Failed to reset post-processing settings.')

    def _update_mode(self):
        """Updates the mode of the camera, which is the bit-wise or between
           exposure mode and expose out mode. It then sets up a small sequence
           so the exposure mode and expose out mode getters will read properly.
           This function should only be called internally whenever either exposure
           setting is changed.

        Side Effect(s):
            - Changes self.__mode
            - Sets up a small sequence so the camera will readout the exposure
              modes correctly with get_param.

        Returns:
            None
        """
        self.__mode = self.__exp_mode | self.__exp_out_mode
        pvc.set_exp_modes(self.__handle, self.__mode)

    def reset_rois(self):
        """Resets the ROI list to default, which is full frame.

        Returns:
            None
        """
        self.__rois = [self.__defaultRoi]

    def set_roi(self, s1, p1, w, h):
        """Configures a ROI for the camera. The default ROI is the full frame. If the default is
           set or only a single ROI is supported, this function will over-write that ROI. Otherwise,
           this function will attempt to append this ROI to the list.

        Returns:
            None
        """

        if w < 1 or h < 1:
            raise ValueError('Width and height must be >= 1')

        s2 = s1 + w - 1
        p2 = p1 + h - 1

        # Check if new ROI is in bounds
        inBounds = 0 <= s1 <= self.sensor_size[0] and \
            0 <= s2 <= self.sensor_size[0] and \
            0 <= p1 <= self.sensor_size[1] and \
            0 <= p2 <= self.sensor_size[1]

        if not inBounds:
            raise ValueError('Could not add ROI. ROI extends beyond limits of frame')

        # Get the total supported ROIs
        roiCount = self.get_param(const.PARAM_ROI_COUNT, const.ATTR_MAX)

        # Check if current ROI list only contains the default
        usingDefaultRoi = False
        if len(self.__rois) == 1:
            usingDefaultRoi = self.__rois[0] == self.__defaultRoi

        newRoi = Camera.RegionOfInterest(s1, s2, self.bin_x, p1, p2, self.bin_y)
        if roiCount == 1 or usingDefaultRoi:
            self.__rois = [newRoi]

        elif len(self.__rois) < roiCount:

            # Check that new ROI doesn't overlap with existing ROIs
            for roi in self.__rois:
                if roi.checkOverlap(newRoi):
                    raise ValueError('Could not add ROI. New ROI overlaps existing ROI')

            self.__rois.append(newRoi)
        else:
            raise ValueError('Could not add ROI. Camera only supports {} rois'.format(roiCount))

    def poll_frame(self, timeout_ms=WAIT_FOREVER, oldestFrame=True, copyData=True):
        """Calls the pvc.get_frame function with the current camera settings.

        Parameter:
            oldestFrame (bool): Selects whether to return the oldest or newest frame. Only the oldest frame will be popped off the underlying queue of frames. (optional).
            copyData (bool): Selects whether to return a copy of the numpy frame which points to a new buffer, or the original numpy frame which points to the
                             buffer used directly by PVCAM. Disabling this copy is not recommended for most situations. Refer to PyVCAM Wrapper.md for more details. (optional).

            None
        Returns:
            A dictionary with the frame containing available meta data and 2D np.array pixel data, frames per second and frame count.
        """

        frame, fps, frame_count = pvc.get_frame(self.__handle, self.__rois, self.__dtype.num, timeout_ms, oldestFrame)

        num_rois = len(frame['pixel_data'])
        if copyData:
            frameTmp = {'pixel_data': [None] * num_rois}
            for roi_index in range(num_rois):
                frameTmp['pixel_data'][roi_index] = np.copy(frame['pixel_data'][roi_index])

            if 'meta_data' in frame.keys():
                frameTmp['meta_data'] = deepcopy(frame['meta_data'])

            frame = frameTmp

        # If using a single ROI, remove list container
        if num_rois == 1:
            frame['pixel_data'] = frame['pixel_data'][0]

        return frame, fps, frame_count

    def get_frame(self, exp_time=None, timeout_ms=WAIT_FOREVER):
        """Calls the pvc.get_frame function with the current camera settings.

        Parameter:
            exp_time (int): The exposure time (optional).
        Returns:
            A 2D np.array containing the pixel data from the captured frame.
        """
        self.start_seq(exp_time=exp_time, num_frames=1)
        frame, fps, frame_count = self.poll_frame(timeout_ms=timeout_ms)
        self.finish()

        return frame['pixel_data']

    def get_sequence(self, num_frames, exp_time=None, timeout_ms=WAIT_FOREVER, interval=None):
        """Calls the pvc.get_frame function with the current camera settings in
            rapid-succession for the specified number of frames

        Parameter:
            num_frames (int): Number of frames to capture in the sequence
            exp_time (int): The exposure time (optional)
            interval (int): The time in milliseconds to wait between captures
        Returns:
            A 3D np.array containing the pixel data from the captured frames.
        """

        if len(self.__rois) > 1:
            raise ValueError('get_sequence does not support multi-roi captures')

        shape = self.__rois[0].shape
        stack = np.empty((num_frames, shape[1], shape[0]), dtype=self.__dtype)

        for i in range(num_frames):
            stack[i] = self.get_frame(exp_time=exp_time, timeout_ms=timeout_ms)

            if isinstance(interval, int):
                time.sleep(interval / 1000)

        return stack

    def get_vtm_sequence(self, time_list, exp_res, num_frames, timeout_ms=WAIT_FOREVER, interval=None):
        """Calls the pvc.get_frame function within a loop, setting vtm expTime
            between each capture.

        Parameter:
            time_list (list of ints): List of vtm timings
            exp_res (int): vtm exposure time resolution (0:mili, 1:micro)
            num_frames (int): Number of frames to capture in the sequence
            interval (int): The time in milliseconds to wait between captures
        Returns:
            A 3D np.array containing the pixel data from the captured sequence.
        """
        old_res = self.exp_res
        self.exp_res = exp_res

        if len(self.__rois) > 1:
            raise ValueError('get_vtm_sequence does not support multi-roi captures')

        shape = self.__rois[0].shape
        stack = np.empty((num_frames, shape[1], shape[0]), dtype=self.__dtype)

        for i in range(num_frames):
            exp_time = time_list[i]
            try:
                self.vtm_exp_time = exp_time
                stack[i] = self.get_frame(exp_time=self.vtm_exp_time, timeout_ms=timeout_ms)
            except Exception:
                raise ValueError('Could not collect vtm frame')

            if isinstance(interval, int):
                time.sleep(interval / 1000)

        self.exp_res = old_res
        return stack

    def start_live(self, exp_time=None, buffer_frame_count=16, stream_to_disk_path=None):
        """Calls the pvc.start_live function to setup a circular buffer acquisition.

        Parameter:
            exp_time (int): The exposure time (optional).
        Returns:
            None
        """
        self._set_dtype()

        if not isinstance(exp_time, int):
            exp_time = self.exp_time

        if isinstance(stream_to_disk_path, str):
            stream_to_disk_path_abs = os.path.abspath(stream_to_disk_path)
            directory, filename = os.path.split(stream_to_disk_path_abs)
            if os.path.exists(directory):
                try:
                    os.remove(filename)
                except OSError:
                    pass
            else:
                raise ValueError('Invalid directory for stream to disk: ' + directory)

        self.__acquisition_mode = 'Live'
        self.__exposure_bytes = pvc.start_live(self.__handle, self.__rois, exp_time, self.__mode, buffer_frame_count, stream_to_disk_path)

    def start_seq(self, exp_time=None, num_frames=1):
        """Calls the pvc.start_seq function to setup a non-circular buffer acquisition.

        Parameter:
            exp_time (int): The exposure time (optional).
        Returns:
            None
        """
        self._set_dtype()

        if not isinstance(exp_time, int):
            exp_time = self.exp_time

        self.__acquisition_mode = 'Sequence'
        self.__exposure_bytes = pvc.start_seq(self.__handle, self.__rois, exp_time, self.__mode, num_frames)

    def finish(self):
        """Ends a previously started live or sequence acquisition.

        Parameter:
            None
        Returns:
            None
        """
        if self.__acquisition_mode == 'Live':
            pvc.stop_live(self.__handle)
        elif self.__acquisition_mode == 'Sequence':
            pvc.finish_seq(self.__handle)

        self.__acquisition_mode = None
        return

    def abort(self):
        """Calls the pvc.abort function that aborts acquisition.

        Parameter:
            None
        Returns:
            None
        """
        return pvc.abort(self.__handle)

    def sw_trigger(self):
        """Performs a SW trigger. This trigger behaves analogously to a HW external trigger. Will throw an exception if trigger fails.

        Parameter:
            None
        Returns:
            None
        """

        pvc.sw_trigger(self.__handle)

    def set_post_processing_param(self, feature_name, param_name, value):
        """Sets the value of a post processing parameter.

        Parameter:
            Feature name and parameter name as specified in post_processing_table
        Returns:
            None
        """

        if feature_name in self.__post_processing_table.keys():
            if param_name in self.__post_processing_table[feature_name].keys():
                pp_param = self.__post_processing_table[feature_name][param_name]

                if pp_param['param_min'] <= value <= pp_param['param_max']:
                    self.set_param(const.PARAM_PP_INDEX, pp_param['feature_index'])
                    self.set_param(const.PARAM_PP_PARAM_INDEX, pp_param['param_index'])
                    self.set_param(const.PARAM_PP_PARAM, value)
                else:
                    raise AttributeError('Could not set post processing param. Value ' + str(value) + ' out of range (' + str(pp_param['param_min']) + ', ' + str(pp_param['param_max']) + ')')
            else:
                raise AttributeError('Could not set post processing param. param_name not found')
        else:
            raise AttributeError('Could not set post processing param. feature_name not found')

    def get_post_processing_param(self, feature_name, param_name):
        """Gets the current value of a post processing parameter.

        Parameter:
            Feature name and parameter name as specified in post_processing_table
        Returns:
            Value of specified post processing parameter
        """

        if feature_name in self.__post_processing_table.keys():
            if param_name in self.__post_processing_table[feature_name].keys():
                pp_param = self.__post_processing_table[feature_name][param_name]
                self.set_param(const.PARAM_PP_INDEX, pp_param['feature_index'])
                self.set_param(const.PARAM_PP_PARAM_INDEX, pp_param['param_index'])
                return self.get_param(const.PARAM_PP_PARAM)
            else:
                raise AttributeError('Could not set post processing param. param_name not found')
        else:
            raise AttributeError('Could not set post processing param. feature_name not found')

    def _set_dtype(self):
        bit_depth = self.bit_depth
        bytes_per_pixel = int(np.ceil(bit_depth / 8))
        dtypeKind = 'u{}'.format(bytes_per_pixel)
        self.__dtype = np.dtype(dtypeKind)

    ### Getters/Setters below ###

    @property
    def handle(self):
        return self.__handle

    @property
    def is_open(self):
        return self.__is_open

    @property
    def name(self):
        return self.__name

    @property
    def post_processing_table(self):
        return self.__post_processing_table

    @property
    def port_speed_gain_table(self):
        return self.__port_speed_gain_table

    @property
    def centroids_modes(self):
        return self.__centroids_modes

    @property
    def clear_modes(self):
        return self.__clear_modes

    @property
    def exp_modes(self):
        return self.__exp_modes

    @property
    def exp_out_modes(self):
        return self.__exp_out_modes

    @property
    def exp_resolutions(self):
        return self.__exp_resolutions

    @property
    def prog_scan_modes(self):
        return self.__prog_scan_modes

    @property
    def prog_scan_dirs(self):
        return self.__prog_scan_dirs

    @property
    def driver_version(self):
        dd_ver = self.get_param(const.PARAM_DD_VERSION)
        # The device driver version is returned as a highly formatted 16 bit
        # integer where the first 8 bits are the major version, bits 9-12 are
        # the minor version, and bits 13-16 are the build number. Uses of masks
        # and bit shifts are required to extract the full version number.
        return '{}.{}.{}'.format(dd_ver & 0xff00 >> 8,
                                 dd_ver & 0x00f0 >> 4,
                                 dd_ver & 0x000f)

    @property
    def cam_fw(self):
        return pvc.get_cam_fw_version(self.__handle)

    @property
    def chip_name(self):
        return self.get_param(const.PARAM_CHIP_NAME)

    @property
    def sensor_size(self):
        return (self.get_param(const.PARAM_SER_SIZE),
                self.get_param(const.PARAM_PAR_SIZE))

    @property
    def serial_no(self):
        # HACK: cytocam fix for messed up serial numbers
        try:
            serial_no = self.get_param(const.PARAM_HEAD_SER_NUM_ALPHA)
            return serial_no
        except:
            return 'N/A'

    @property
    def bit_depth(self):
        return self.get_param(const.PARAM_BIT_DEPTH)

    @property
    def pix_time(self):
        return self.get_param(const.PARAM_PIX_TIME)

    @property
    def readout_port(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_READOUT_PORT)

    @readout_port.setter
    def readout_port(self, value):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        num_ports = self.get_param(const.PARAM_READOUT_PORT, const.ATTR_COUNT)

        if value >= num_ports:
            raise ValueError('{} only supports '
                             '{} readout ports.'.format(self, num_ports))
        self.set_param(const.PARAM_READOUT_PORT, value)

    @property
    def speed_table_index(self):
        return self.get_param(const.PARAM_SPDTAB_INDEX)

    @speed_table_index.setter
    def speed_table_index(self, value):
        num_entries = self.get_param(const.PARAM_SPDTAB_INDEX,
                                     const.ATTR_COUNT)
        if value >= num_entries:
            raise ValueError('{} only supports '
                             '{} speed entries'.format(self, num_entries))
        self.set_param(const.PARAM_SPDTAB_INDEX, value)

    @property
    def trigger_table(self):
        # Returns a dictionary containing information about the last capture.
        # Note some features are camera specific.

        if self.exp_res == 1:
            exp = str(self.last_exp_time) + ' μs'
        else:
            exp = str(self.last_exp_time) + ' ms'

        try:
            read = str(self.readout_time) + ' μs'
        except AttributeError:
            read = 'N/A'

        # If the camera has clear time, then it has pre and post trigger delays
        try:
            clear = str(self.clear_time) + ' ns'
            pre = str(self.pre_trigger_delay) + ' ns'
            post = str(self.post_trigger_delay) + ' ns'
        except:
            clear = 'N/A'
            pre = 'N/A'
            post = 'N/A'

        return {'Exposure Time': exp,
                'Readout Time': read,
                'Clear Time': clear,
                'Pre-trigger Delay': pre,
                'Post-trigger Delay': post}

    @property
    def adc_offset(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_ADC_OFFSET)

    @property
    def gain(self):
        return self.get_param(const.PARAM_GAIN_INDEX)

    @gain.setter
    def gain(self, value):
        min_gain = self.get_param(const.PARAM_GAIN_INDEX, const.ATTR_MIN)
        max_gain = self.get_param(const.PARAM_GAIN_INDEX, const.ATTR_MAX)
        if not (min_gain <= value <= max_gain):
            raise ValueError("Invalid value: {} - {} only supports gain "
                             "indicies from {} - {}.".format(value, self, min_gain, max_gain))
        self.set_param(const.PARAM_GAIN_INDEX, value)

    @property
    def binning(self):
        return self.bin_x, self.bin_y

    @binning.setter
    def binning(self, value):
        if isinstance(value, tuple):
            self.bin_x = value[0]
            self.bin_y = value[1]
        else:
            self.binning = (value, value)

    @property
    def bin_x(self):
        return self.__rois[0].sbin

    @bin_x.setter
    def bin_x(self, value):
        # Will raise ValueError if incompatible binning is set
        if value in self.read_enum(const.PARAM_BINNING_SER).values():
            for roi in self.__rois:
                roi.sbin = value
            return

        raise ValueError('{} only supports {} binnings'.format(self,
                                                               self.read_enum(const.PARAM_BINNING_SER).items()))

    @property
    def bin_y(self):
        return self.__rois[0].pbin

    @bin_y.setter
    def bin_y(self, value):
        # Will raise ValueError if incompatible binning is set
        if value in self.read_enum(const.PARAM_BINNING_PAR).values():
            for roi in self.__rois:
                roi.pbin = value
            return

        raise ValueError('{} only supports {} binnings'.format(self,
                                                               self.read_enum(const.PARAM_BINNING_PAR).items()))

    def shape(self, roi_index=0):
        return self.__rois[roi_index].shape

    @property
    def last_exp_time(self):
        return self.get_param(const.PARAM_EXPOSURE_TIME)

    @property
    def exp_res(self):
        return self.get_param(const.PARAM_EXP_RES)

    @exp_res.setter
    def exp_res(self, keyOrValue):
        # Will raise ValueError if provided with an unrecognized key.
        value = self.__exp_resolutions[keyOrValue] if isinstance(keyOrValue, str) else keyOrValue

        # Verify value is in range by attempting to look-up the key
        key = self.__exp_resolutions[value]

        self.set_param(const.PARAM_EXP_RES, value)

    @property
    def exp_res_index(self):
        return self.get_param(const.PARAM_EXP_RES_INDEX)

    @property
    def exp_time(self):
        # TODO: Testing
        return self.__exp_time

    @exp_time.setter
    def exp_time(self, value):
        min_exp_time = self.get_param(const.PARAM_EXPOSURE_TIME, const.ATTR_MIN)
        max_exp_time = self.get_param(const.PARAM_EXPOSURE_TIME, const.ATTR_MAX)

        if not value in range(min_exp_time, max_exp_time + 1):
            raise ValueError("Invalid value: {} - {} only supports exposure "
                             "times between {} and {}".format(value, self,
                                                              min_exp_time,
                                                              max_exp_time))
        self.__exp_time = value

    @property
    def exp_mode(self):
        return self.get_param(const.PARAM_EXPOSURE_MODE)

    @exp_mode.setter
    def exp_mode(self, keyOrValue):
        # Will raise ValueError if provided with an unrecognized key.
        self.__exp_mode = self.__exp_modes[keyOrValue] if isinstance(keyOrValue, str) else keyOrValue

        # Verify value is in range by attempting to look-up the key
        key = self.__exp_modes[self.__exp_mode]

        self._update_mode()

    @property
    def exp_out_mode(self):
        return self.get_param(const.PARAM_EXPOSE_OUT_MODE)

    @exp_out_mode.setter
    def exp_out_mode(self, keyOrValue):
        # Will raise ValueError if provided with an unrecognized key.
        self.__exp_out_mode = self.__exp_out_modes[keyOrValue] if isinstance(keyOrValue, str) else keyOrValue

        # Verify value is in range by attempting to look-up the key
        key = self.__exp_out_modes[self.__exp_out_mode]

        self._update_mode()

    @property
    def vtm_exp_time(self):
        return self.get_param(const.PARAM_EXP_TIME)

    @vtm_exp_time.setter
    def vtm_exp_time(self, value):
        min_exp_time = self.get_param(const.PARAM_EXPOSURE_TIME, const.ATTR_MIN)
        max_exp_time = self.get_param(const.PARAM_EXPOSURE_TIME, const.ATTR_MAX)

        if not value in range(min_exp_time, max_exp_time + 1):
            raise ValueError("Invalid value: {} - {} only supports exposure "
                             "times between {} and {}".format(value, self,
                                                              min_exp_time,
                                                              max_exp_time))
        self.set_param(const.PARAM_EXP_TIME, value)

    @property
    def clear_mode(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_CLEAR_MODE)

    @clear_mode.setter
    def clear_mode(self, keyOrValue):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting. Will raise ValueError if provided with an unrecognized key.
        value = self.__clear_modes[keyOrValue] if isinstance(keyOrValue, str) else keyOrValue

        # Verify value is in range by attempting to look-up the key
        key = self.__clear_modes[value]

        self.set_param(const.PARAM_CLEAR_MODE, value)

    @property
    def temp(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_TEMP) / 100.0

    @property
    def temp_setpoint(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_TEMP_SETPOINT) / 100.0

    @temp_setpoint.setter
    def temp_setpoint(self, value):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        try:
            self.set_param(const.PARAM_TEMP_SETPOINT, value)
        except RuntimeError:
            min_temp = self.get_param(const.PARAM_TEMP_SETPOINT, const.ATTR_MIN)
            max_temp = self.get_param(const.PARAM_TEMP_SETPOINT, const.ATTR_MAX)
            raise ValueError("Invalid temp {} : Valid temps are in the range {} "
                             "- {}.".format(value, min_temp, max_temp))

    @property
    def readout_time(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_READOUT_TIME)

    @property
    def clear_time(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_CLEARING_TIME)

    @property
    def pre_trigger_delay(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_PRE_TRIGGER_DELAY)

    @property
    def post_trigger_delay(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_POST_TRIGGER_DELAY)

    @property
    def centroids_mode(self):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_CENTROIDS_MODE)

    @centroids_mode.setter
    def centroids_mode(self, keyOrValue):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting. Will raise ValueError if
        # provided with an unrecognized key
        value = self.__centroids_modes[keyOrValue] if isinstance(keyOrValue, str) else keyOrValue

        # Verify value is in range by attempting to look-up the key
        key = self.__centroids_modes[value]

        self.set_param(const.PARAM_CENTROIDS_MODE, value)

    @property
    def scan_line_time(self):
        return self.get_param(const.PARAM_SCAN_LINE_TIME)

    @property
    def prog_scan_mode(self):
        # Camera specific setting: Will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_SCAN_MODE)

    @prog_scan_mode.setter
    def prog_scan_mode(self, keyOrValue):
        # Camera specific setting: will raise AttributeError if called with a
        # camera that does not support this setting. Will raise ValueError if
        # provided with an unrecognized key
        value = self.__prog_scan_modes[keyOrValue] if isinstance(keyOrValue, str) else keyOrValue

        # Verify value is in range by attempting to look-up the key
        key = self.__prog_scan_modes[value]

        self.set_param(const.PARAM_SCAN_MODE, value)

    @property
    def prog_scan_dir(self):
        # Camera specific setting: Will raise AttributeError if called with a
        # camera that does not support this setting.
        return self.get_param(const.PARAM_SCAN_DIRECTION)

    @prog_scan_dir.setter
    def prog_scan_dir(self, keyOrValue):
        # Camera specific setting. Will raise AttributeError if called with a
        # camera that does not support this setting. Will raise ValueError if
        # provided with an unrecognized key or value
        value = self.__prog_scan_dirs[keyOrValue] if isinstance(keyOrValue, str) else keyOrValue

        # Verify value is in range by attempting to look-up the key
        key = self.__prog_scan_dirs[value]

        self.set_param(const.PARAM_SCAN_DIRECTION, value)

    @property
    def prog_scan_dir_reset(self):
        return self.get_param(const.PARAM_SCAN_DIRECTION_RESET)

    @prog_scan_dir_reset.setter
    def prog_scan_dir_reset(self, value):
        self.set_param(const.PARAM_SCAN_DIRECTION_RESET, value)

    @property
    def prog_scan_line_delay(self):
        return self.get_param(const.PARAM_SCAN_LINE_DELAY)

    @prog_scan_line_delay.setter
    def prog_scan_line_delay(self, value):
        self.set_param(const.PARAM_SCAN_LINE_DELAY, value)

    @property
    def prog_scan_width(self):
        return self.get_param(const.PARAM_SCAN_WIDTH)

    @prog_scan_width.setter
    def prog_scan_width(self, value):
        self.set_param(const.PARAM_SCAN_WIDTH, value)

    @property
    def meta_data_enabled(self):
        return self.get_param(const.PARAM_METADATA_ENABLED)

    @meta_data_enabled.setter
    def meta_data_enabled(self, value):
        self.set_param(const.PARAM_METADATA_ENABLED, value)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import os
import threading
import ctypes
import sys
from collections import deque
import logging
import traceback

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class SelfLockThread(threading.Thread):
    """A custom thread class with self-locking capabilities.

    This class extends the functionality of the Python threading.Thread class by
    providing the ability to lock and unlock the thread explicitly. It is useful
    in scenarios where thread synchronization and control are required.

    Note
    ----
    - This class provides additional control over thread execution using the
      `wait()` and `unlock()` methods.
    - It allows checking whether the thread is currently locked using the `isLocked()`
      method.
    """

    def __init__(
        self, group=None, target=None, name=None, args=(), kwargs={}, *, daemon=None
    ):
        """Initialize the SelfLockThread.

        Parameters
        ----------
        group : ThreadGroup, optional
            The thread group, by default None
        target : callable, optional
            The target function of the thread, by default None
        name : str, optional
            The name of the thread, by default None
        args : tuple, optional
            The arguments of the target function, by default ()
        kwargs : dict, optional
            The keyword arguments of the target function, by default {}
        daemon : bool, optional
            Whether the thread is daemon, by default None
        """
        super().__init__(group, target, name, args, kwargs, daemon=daemon)
        #: threading.Lock: The lock of the thread.
        self.selfLock = threading.Lock()
        # lock itself
        self.selfLock.acquire()

    def run(self):
        """Run the thread."""
        self._kwargs["thread"] = self
        if self._target:
            try:
                self._target(*self._args, **self._kwargs)
            except Exception as e:
                print(
                    f"{self.name} thread ended because of exception!: {e}",
                    traceback.format_exc(),
                )
                logger.debug(
                    f"{self.name} thread ended because of exception!: {e}",
                    traceback.format_exc(),
                )
            finally:
                pass

    def wait(self):
        """Wait for the thread to finish."""
        self.selfLock.acquire()

    def unlock(self):
        """Unlock the thread."""
        if self.selfLock.locked():
            self.selfLock.release()

    def isLocked(self):
        """Check if the thread is locked.

        Returns
        -------
        bool
            Whether the thread is locked.
        """
        return self.selfLock.locked()


class SynchronizedThreadPool:
    """
    A custom thread pool with synchronization and control features.

    This class provides a thread pool for managing threads associated with different
    resources. It allows registering resources, creating threads, wrapping thread tasks,
    removing threads, and other thread-related operations with explicit control and
    synchronization.

    Note
    ----
    - This class provides explicit control over thread creation, removal,
    and synchronization.
    - It allows managing threads associated with different resources efficiently.
    - The `clear` method moves threads to the `toDeleteList` for later deletion.

    """

    def __init__(self):
        """Initialize the SynchronizedThreadPool."""

        #: dict: The resources of the thread pool.
        self.resources = {}
        #: dict: The toDeleteList of the thread pool.
        self.toDeleteList = {}

    def registerResource(self, resourceName):
        """Register a resource to the pool.

        Parameters
        ----------
        resourceName : str
            The name of the resource.
        """
        if resourceName not in self.resources:
            self.resources[resourceName] = ThreadWaitlist()

    def createThread(
        self,
        resourceName,
        target,
        args=(),
        kwargs={},
        *,
        callback=None,
        cbArgs=(),
        cbKargs={},
    ):
        """Create a thread and add it to the waitlist of the resource.

        Parameters
        ----------
        resourceName : str
            The name of the resource.
        target : callable
            The target function of the thread.
        args : tuple, optional
            The arguments of the target function, by default ()
        kwargs : dict, optional
            The keyword arguments of the target function, by default {}
        callback : callable, optional
            The callback function of the thread, by default None
        cbArgs : tuple, optional
            The arguments of the callback function, by default ()
        cbKargs : dict, optional
            The keyword arguments of the callback function, by default {}

        Returns
        -------
        SelfLockThread
            The created thread.
        """

        if resourceName not in self.resources:
            self.registerResource(resourceName)
        task = self.threadTaskWrapping(
            resourceName, target, callback=callback, cbArgs=cbArgs, cbKargs=cbKargs
        )
        taskThread = SelfLockThread(None, task, resourceName, args, kwargs, daemon=True)
        taskThread.start()
        return taskThread

    def threadTaskWrapping(
        self, resourceName, target, *, callback=None, cbArgs=(), cbKargs={}
    ):
        """Wrap the target function of the thread.

        Parameters
        ----------
        resourceName : str
            The name of the resource.
        target : callable
            The target function of the thread.
        callback : callable, optional
            The callback function of the thread, by default None
        cbArgs : tuple, optional
            The arguments of the callback function, by default ()
        cbKargs : dict, optional
            The keyword arguments of the callback function, by default {}

        Returns
        -------
        callable
            The wrapped target function.
        """

        def func(*args, **kwargs):
            thread = kwargs.get("thread", None)
            if not thread:
                return
            del kwargs["thread"]
            # add thread to the waitlist of the resource
            with self.resources[resourceName] as resource:
                resource.waitlist.append(thread)
                if len(resource.waitlist) == 1:
                    thread.unlock()
            # wait for it's turn
            thread.wait()
            # run itself
            if callable(target):
                try:
                    target(*args, **kwargs)
                except Exception as e:
                    print(
                        threading.current_thread().name,
                        "thread exception happened!",
                        e,
                        traceback.format_exc(),
                    )
                    logger.debug(
                        threading.current_thread().name,
                        "thread exception happened!",
                        e,
                        traceback.format_exc(),
                    )

            # wake up next thread if any
            with self.resources[resourceName] as resource:
                resource.waitlist.popleft()
                if len(resource.waitlist) > 0:
                    resource.waitlist[0].unlock()
            # run callback
            if callback:
                callback(*cbArgs, **cbKargs)

        return func

    def removeThread(self, resourceName, taskThread):
        """Remove a thread from the waitlist of the resource.

        Parameters
        ----------
        resourceName : str
            The name of the resource.
        taskThread : SelfLockThread
            The thread to remove.

        Returns
        -------
        bool
            Whether the thread is removed.
        """
        # can only remove waiting threads
        # do not remove running threading
        # if no such resource
        if resourceName not in self.resources:
            return False
        with self.resources[resourceName] as resource:
            # if it is the current running one, kill it and wake up next thread if any
            if resource.waitlist[0] != taskThread:
                # add the thread to toDeleteList
                self.moveToDelete(resourceName, taskThread)
                resource.waitlist.remove(taskThread)
                return

    def moveToDelete(self, resourceName, taskThread):
        """Move a thread to the toDeleteList.

        Parameters
        ----------
        resourceName : str
            The name of the resource.
        taskThread : SelfLockThread
            The thread to move.
        """
        if resourceName not in self.toDeleteList:
            self.toDeleteList[resourceName] = ThreadWaitlist()
        with self.toDeleteList[resourceName] as temp:
            temp.waitlist.append(taskThread)

    def getRunningThread(self, resourceName):
        """Get the running thread of the resource.

        Parameters
        ----------
        resourceName : str
            The name of the resource.

        Returns
        -------
        SelfLockThread
            The running thread.
        """
        if (
            resourceName not in self.resources
            or len(self.resources[resourceName].waitlist) < 1
        ):
            return None
        return self.resources[resourceName].waitlist[0]

    def clear(self):
        """Clear all the threads in the pool."""

        # move all the threads except first one to toDeleteList
        sys.settrace(self.globaltrace)
        for resourceName in self.resources:
            with self.resources[resourceName] as temp:
                headThread = temp.waitlist[0] if len(temp.waitlist) > 0 else None
                for i in range(1, len(temp.waitlist)):
                    self.moveToDelete(resourceName, temp.waitlist[i])
            if headThread:
                self._raiseError(headThread.native_id)
            self.killThreadInList(resourceName, self.toDeleteList)

    def killThreadInList(self, resourceName, threadList):
        """Kill all the threads in the threadList.

        Parameters
        ----------
        resourceName : str
            The name of the resource.
        threadList : dict
            The threadList to kill.
        """
        sys.settrace(self.globaltrace)
        # remove all the threads in threadList
        if resourceName in threadList:
            with threadList[resourceName] as temp:
                while temp.waitlist:
                    thread = temp.waitlist[0]
                    thread.unlock()
                    self._raiseError(thread.native_id)
                    thread.join(1)
                    if not thread.is_alive():
                        # move it from the list
                        temp.waitlist.popleft()
                    else:
                        thread.wait()
                        # move the thread to the end of the list
                        temp.waitlist.popleft()
                        temp.waitlist.append(thread)

    def globaltrace(self, frame, event, arg):
        """Global trace function.

        Parameters
        ----------
        frame : frame
            The frame of the thread.
        event : str
            The event of the thread.
        arg : any
            The argument of the thread.

        Returns
        -------
        callable
            The local trace function if the event is "call", otherwise None.
        """

        if event == "call":
            return self.localtrace
        else:
            return None

    def localtrace(self, frame, event, arg):
        """Local trace function.

        Parameters
        ----------
        frame : frame
            The frame of the thread.
        event : str
            The event of the thread.
        arg : any
            The argument of the thread.

        Returns
        -------
        callable
            The local trace function if the event is "exception", otherwise None.
        """

        if event == "exception":
            print("****in local trace: exception stops the thread")
            logger.debug("****in local trace: exception stops the thread")
            if os.getenv("GITHUB_ACTIONS") == "true":
                return
            raise SystemExit()
        return self.localtrace

    def _raiseError(self, threadId):
        """Raise error in the thread.

        Parameters
        ----------
        threadId : int
            The id of the thread.
        """

        ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_ulong(threadId), ctypes.py_object(SystemExit)
        )


class ThreadWaitlist:
    """A custom thread waitlist with synchronization capabilities."""

    def __init__(self):
        """Initialize the ThreadWaitlist."""
        #: threading.Lock: The lock of the waitlist.
        self.waitlistLock = threading.Lock()
        #: deque: The waitlist.
        self.waitlist = deque()

    def __enter__(self):
        """Enter the context.

        Returns
        -------
        ThreadWaitlist
            The waitlist.
        """
        self.waitlistLock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context.

        Parameters
        ----------
        exc_type : type
            The type of the exception.
        exc_val : any
            The value of the exception.
        exc_tb : traceback
            The traceback of the exception.
        """
        self.waitlistLock.release()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import tkinter as tk
from tkinter import ttk
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedSpinbox
import logging

p = __name__.split(".")[1]
logger = logging.getLogger(p)


class WaveformParameterPopupWindow:
    """Popup window with waveform parameters for galvos, remote focusing, etc."""

    def __init__(self, root, configuration_controller, *args, **kwargs):
        # Creating popup window with this name and size/placement, PopUp is a
        # Toplevel window
        #: PopUp: The popup window
        self.popup = PopUp(
            root, "Waveform Parameter Settings", "+320+180", top=False, transient=False
        )

        #: configuration_controller: The configuration controller
        self.configuration_controller = configuration_controller

        # Storing the content frame of the popup, this will be the parent of
        # the widgets
        content_frame = self.popup.get_frame()
        content_frame.columnconfigure(0, pad=5)
        content_frame.columnconfigure(1, pad=5)
        content_frame.rowconfigure(0, pad=5)
        content_frame.rowconfigure(1, pad=5)
        content_frame.rowconfigure(2, pad=5)

        # Formatting
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        #: dict: Dictionary for all the variables
        self.inputs = {}
        #: dict: Dictionary for all the buttons
        self.buttons = {}

        # Frames for widgets
        #: ttk.Frame: Frame for mode and magnification
        self.mode_mag_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))
        #: ttk.Frame: Frame for saving waveform parameters
        self.save_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))
        #: ttk.Frame: Frame for laser parameters
        self.laser_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))
        #: ttk.Frame: Frame for high/low resolution
        self.high_low_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))

        # Griding Frames
        self.mode_mag_frame.grid(row=0, column=0, sticky=tk.NSEW)
        self.save_frame.grid(row=0, column=1, sticky=tk.NSEW)
        self.laser_frame.grid(row=2, column=0, columnspan=2, sticky=tk.NSEW)
        self.high_low_frame.grid(row=3, column=0, columnspan=2, sticky=tk.NSEW)

        # Filling Frames with widgets
        # Mode/Mag Frame
        self.inputs["Mode"] = LabelInput(
            parent=self.mode_mag_frame,
            label="Mode",
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            label_args={"padding": (2, 5, 48, 0)},
        )
        self.inputs["Mode"].grid(row=0, column=0)
        self.inputs["Mode"].state(["readonly"])

        self.inputs["Mag"] = LabelInput(
            parent=self.mode_mag_frame,
            label="Magnification",
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            label_args={"padding": (2, 5, 5, 0)},
        )
        self.inputs["Mag"].grid(row=1, column=0)
        self.inputs["Mag"].state(["readonly"])

        # Save Waveform Parameters Frame
        self.buttons["Save"] = ttk.Button(self.save_frame, text="Save Configuration")
        self.buttons["Save"].grid(
            row=0, column=0, sticky=tk.NSEW, padx=(5, 0), pady=(0, 0)
        )

        # Toggle Waveform Button
        self.buttons["toggle_waveform_button"] = ttk.Button(
            self.save_frame,
            text="Disable Waveforms",
        )
        self.buttons["toggle_waveform_button"].grid(
            row=1, column=0, sticky=tk.NSEW, padx=(5, 0), pady=(0, 0)
        )

        # Laser Frame
        laser_labels = self.configuration_controller.lasers_info
        title_labels = ["Laser", "Amplitude", "Offset"]
        # Loop for widgets
        for i in range(3):
            # Title labels
            title = ttk.Label(
                self.laser_frame, text=title_labels[i], padding=(2, 5, 0, 0)
            )
            title.grid(row=0, column=i, sticky=tk.NSEW, padx=(0, 5))
        for i, label in enumerate(laser_labels):
            # Laser labels
            laser = ttk.Label(
                self.laser_frame, text=laser_labels[i], padding=(2, 5, 0, 0)
            )
            laser.grid(row=i + 1, column=0, sticky=tk.NSEW)

            # Entry Widgets
            self.inputs[laser_labels[i] + " Amp"] = LabelInput(
                parent=self.laser_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
            )

            self.inputs[laser_labels[i] + " Amp"].grid(
                row=i + 1, column=1, sticky=tk.NSEW, pady=(20, 0), padx=(0, 5)
            )

            self.inputs[laser_labels[i] + " Off"] = LabelInput(
                parent=self.laser_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
            )

            self.inputs[laser_labels[i] + " Off"].grid(
                row=i + 1, column=2, sticky=tk.NSEW, pady=(20, 0)
            )

        galvo_labels = list(
            map(lambda i: f"Galvo {i}", range(self.configuration_controller.galvo_num))
        )

        prev = len(laser_labels)

        for i, label in enumerate(galvo_labels):
            galvo = ttk.Label(
                self.laser_frame, text=galvo_labels[i], padding=(2, 5, 0, 0)
            )

            galvo.grid(row=prev + 1, column=0, sticky=tk.NSEW)

            self.inputs[galvo_labels[i] + " Amp"] = LabelInput(
                parent=self.laser_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
            )

            self.inputs[galvo_labels[i] + " Amp"].grid(
                row=prev + 1, column=1, sticky=tk.NSEW, pady=(20, 0), padx=(0, 5)
            )

            self.inputs[galvo_labels[i] + " Off"] = LabelInput(
                parent=self.laser_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
            )

            self.inputs[galvo_labels[i] + " Off"].grid(
                row=prev + 1, column=2, sticky=tk.NSEW, pady=(20, 0)
            )

            galvo_freq = ttk.Label(
                self.laser_frame,
                text=galvo_labels[i] + " Freq (Hz)",
                padding=(2, 5, 0, 0),
            )

            galvo_freq.grid(row=prev + 2, column=0, sticky=tk.NSEW)

            self.inputs[galvo_labels[i] + " Freq"] = LabelInput(
                parent=self.laser_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
                input_args={"from_": 0, "to": 1000, "increment": 0.1},
            )

            self.inputs[galvo_labels[i] + " Freq"].grid(
                row=prev + 2, column=1, sticky=tk.NSEW, pady=(20, 0)
            )

            # Button for automatic estimate of galvo frequency
            self.buttons[galvo_labels[i] + " Freq"] = ttk.Button(
                self.laser_frame,
                text="Estimate Frequency",
            )

            self.buttons[galvo_labels[i] + " Freq"].grid(
                row=prev + 2, column=2, sticky=tk.NSEW, pady=(20, 0)
            )

            prev = prev + 2

        # High/Low Resolution
        hi_lo_labels = ["Percent Delay", "Percent Smoothing", "Settle Duration (ms)"]
        dict_labels = ["Delay", "Smoothing", "Duty"]
        for i in range(3):
            self.inputs[dict_labels[i]] = LabelInput(
                parent=self.high_low_frame,
                input_class=ValidatedSpinbox,
                label=hi_lo_labels[i],
                input_var=tk.StringVar(),
                label_args={"padding": (2, 5, 5, 0)},
                input_args={"from_": 0, "to": 100, "increment": 0.1},
            )
            self.inputs[dict_labels[i]].grid(
                row=i, column=0, sticky=tk.NSEW, padx=(2, 5)
            )

        # Padding Entry Widgets
        self.inputs["Delay"].pad_input(60, 0, 0, 0)
        self.inputs["Smoothing"].pad_input(30, 0, 0, 0)
        self.inputs["Duty"].pad_input(25, 0, 0, 0)

    # Getters
    def get_variables(self):
        """Returns the variables of the inputs

        This function returns a dictionary of all the variables
        that are tied to each widget name.
        The key is the widget name, value is the variable associated.

        Returns
        -------
        dict
            Dictionary of all the variables tied to each widget name

        Examples
        --------
        >>> self.get_variables()
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self):
        """Returns the widgets

        This function returns the dictionary that holds the input widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        dict
            Dictionary of all the widgets

        Examples
        --------
        >>> self.get_widgets()
        """
        return self.inputs

    def get_buttons(self):
        """Returns the buttons

        This function returns the dictionary that holds the buttons.
        The key is the button name, value is the button.

        Returns
        -------
        dict
            Dictionary of all the buttons

        Examples
        --------
        >>> self.get_buttons()
        """
        return self.buttons



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import os
from pathlib import Path
from tkinter import filedialog

# Third party imports
import tifffile

# Local application imports
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.config import get_navigate_path
from navigate.model.analysis.camera import compute_scmos_offset_and_variance_map


class CameraMapSettingPopupController(GUIController):
    """Controller for the camera map setting popup."""

    def __init__(self, view, parent_controller=None):
        """Controller for the camera map setting popup.

        Popup used to generate offset and variance maps of a camera from a
        series of dark frames. The maps are useful for estimating signal-
        to-noise ratios.

        Parameters
        ----------
        view : navigate.view.sub_views.camera_map_setting_popup.CameraMapSettingPopup
            View for the camera map setting popup.
        parent_controller : navigate.controller.main_controller.MainController
            Parent controller of this controller.
        """
        super().__init__(view, parent_controller)
        #: str: Path to the camera maps directory.
        self.map_path = os.path.join(get_navigate_path(), "camera_maps")
        #: np.ndarray: Offset map.
        self.off = None
        #: np.ndarray: Variance map.
        self.var = None

        self.view.open_btn.configure(command=self.open_frames)
        self.view.map_btn.configure(command=self.create_camera_maps)

        menu = self.get_camera_menu()
        self.view.inputs["camera"].set_menu(menu[0], *menu)

    def showup(self):
        """Show the popup.

        This method is called by the parent controller.

        Example
        -------
        >>> self.parent_controller.show_popup("camera_map_setting_popup")
        """
        self.view.showup()

    def get_camera_menu(self):
        """Return list of serial numbers of available cameras.

        Returns
        -------
        menu : list
            Serial numbers of available cameras.
        """
        menu = ["synthetic"]
        for v in self.parent_controller.configuration["configuration"][
            "microscopes"
        ].values():
            menu.append(v["camera"]["hardware"]["serial_number"])
        return menu

    def open_frames(self):
        """Open a file dialog to select a series of dark frames."""
        filename = filedialog.askopenfilename(
            defaultextension=".tif", filetypes=[("tiff files", "*.tif *.tiff")]
        )
        if not filename:
            return
        self.view.file_name.set(filename)

    def create_camera_maps(self):
        """Create offset and variance maps from a series of dark frames."""
        # TODO: This should not be in the controller logic.
        image_name = self.view.file_name.get()
        im = tifffile.imread(image_name)
        self.off, self.var = compute_scmos_offset_and_variance_map(im)

        self.display_plot()

        save_dir = self.map_path
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        camera_name = self.view.camera.get()
        save_name_offset = f"{camera_name}_off.tiff"
        save_name_variance = f"{camera_name}_var.tiff"
        save_path_offset = Path(save_dir).joinpath(save_name_offset)
        save_path_variance = Path(save_dir).joinpath(save_name_variance)

        tifffile.imsave(save_path_offset, self.off)
        tifffile.imsave(save_path_variance, self.var)

    def display_plot(self):
        """Display the offset and variance maps."""
        for ax in self.view.axs:
            ax.clear()

        self.view.axs[0].hist(self.off.ravel(), bins=range(0, 2**8))
        self.view.axs[1].hist(self.var.ravel() * 0.47 * 0.47, bins=range(0, 2**8))

        self.view.axs[0].set_xlabel("Offset (counts)")
        self.view.axs[0].set_ylabel("Frequency")
        self.view.axs[0].set_yscale("log")
        self.view.axs[1].set_xlabel("Variance (counts$^2$)")
        self.view.axs[1].set_ylabel("Frequency")
        self.view.axs[1].set_yscale("log")
        self.view.fig.tight_layout()

        self.view.fig.canvas.draw_idle()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard library imports
import tkinter as tk
from tkinter import ttk

# Third party imports
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

# Local application imports
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput

import logging

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ScrollFrame(ttk.Frame):
    """Scrollable Frame"""

    def __init__(self, parent):
        """Initialize ScrollFrame

        Parameters
        ----------
        parent : tk.Frame
            Parent frame
        """
        tk.Frame.__init__(self, parent)
        #: tk.Canvas: Canvas
        self.canvas = tk.Canvas(self, borderwidth=1)

        #: tk.Frame: Frame
        self.frame = tk.Frame(self.canvas)

        #: tk.Scrollbar: Vertical Scrollbar
        self.vsb = tk.Scrollbar(self, orient="vertical", command=self.canvas.yview)

        self.canvas.configure(yscrollcommand=self.vsb.set, width=200)
        self.vsb.pack(side="right", fill="y")
        self.canvas.pack(side="left", fill="y", expand=False)
        self.canvas.create_window(
            0, 0, window=self.frame, anchor="nw", tags="self.frame"
        )
        self.frame.bind("<Configure>", self.onFrameConfigure)

    def onFrameConfigure(self, event):
        """Reset the scroll region to encompass the inner frame.

        Parameters
        ----------
        event : tk.Event
            Event
        """
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))


class AdaptiveOpticsPopup:
    """Adaptive Optics Popup"""

    def __init__(self, root, *args, **kwargs):
        """Initialize AdaptiveOpticsPopup

        Creating popup window with this name and size/placement, PopUp is a
        Toplevel window

        Parameters
        ----------
        root : tk.Tk
            Root window
        """
        #: PopUp: Popup
        self.popup = PopUp(
            root, "Adaptive Optics", "1100x550+320+180", top=False, transient=False
        )

        #: list: List of mode names
        self.mode_names = [
            "Vert. Tilt",
            "Horz. Tilt",
            "Defocus",
            "Vert. Asm.",
            "Oblq. Asm.",
            "Vert. Coma",
            "Horz. Coma",
            "3rd Spherical",
            "Vert. Tre.",
            "Horz. Tre.",
            "Vert. 5th Asm.",
            "Oblq. 5th Asm.",
            "Vert. 5th Coma",
            "Horz. 5th Coma",
            "5th Spherical",
            "Vert. Tetra.",
            "Oblq. Tetra.",
            "Vert. 7th Tre.",
            "Horz. 7th Tre.",
            "Vert. 7th Asm.",
            "Oblq. 7th Asm.",
            "Vert. 7th Coma",
            "Horz. 7th Coma",
            "7th Spherical",
            "Vert. Penta.",
            "Horz. Penta.",
            "Vert. 9th Tetra.",
            "Oblq. 9th Tetra.",
            "Vert. 9th Tre.",
            "Horz. 9th Tre.",
            "Vert. 9th Asm.",
            "Oblq. 9th Asm.",
        ]

        #: int: Number of modes
        self.n_modes = 32  # TODO: Don't hardcode... Get from exp file!

        content_frame = self.popup.get_frame()

        """Creating the widgets for the popup"""
        #: dict: Dictionary of all the variables
        self.inputs = {}

        #: dict: Dictionary of all the armed modes
        self.modes_armed = {}

        #: dict: Dictionary of all the mode labels
        self.mode_labels = {}

        #: ttk.Notebook: Notebook
        self.ao_notebook = ttk.Notebook(master=content_frame)
        self.ao_notebook.grid(row=0, column=2, rowspan=2)

        #: ttk.Frame: Tony Wilson Tab
        self.tab_tw = ttk.Frame(master=self.ao_notebook)
        self.ao_notebook.add(self.tab_tw, text="Tony Wilson")

        #: ttk.Frame: CNN-AO Tab
        self.tab_cnn = ttk.Frame(master=self.ao_notebook)
        self.ao_notebook.add(self.tab_cnn, text="CNN-AO")

        #: ttk.Frame: Tony Wilson Widget Frame
        tw_widget_frame = ttk.Frame(master=self.tab_tw)
        tw_widget_frame.grid(row=0, column=1)

        #: ttk.Label: Label Frames
        self.inputs["iterations"] = LabelInput(
            tw_widget_frame,
            label="Iterations:",
            label_pos="top",
            input_args={"width": 15},
        )
        self.inputs["iterations"].grid(row=0, column=1, pady=5)
        self.inputs["steps"] = LabelInput(
            tw_widget_frame, label="Steps:", label_pos="top", input_args={"width": 15}
        )
        self.inputs["steps"].grid(row=1, column=1, pady=5)
        self.inputs["amplitude"] = LabelInput(
            tw_widget_frame,
            label="Amplitude:",
            label_pos="top",
            input_args={"width": 15},
        )
        self.inputs["amplitude"].grid(row=2, column=1, pady=5)

        tw_start_from_var = tk.StringVar()
        ttk.Label(tw_widget_frame, text="Start from:").grid(row=3, column=1)
        tw_start_from_combo = ttk.Combobox(
            tw_widget_frame, textvariable=tw_start_from_var, width=20
        )
        tw_start_from_combo["values"] = ("flat", "current")
        tw_start_from_combo.state(["readonly"])
        tw_start_from_combo.grid(row=4, column=1, pady=5)
        tw_start_from_combo.current(0)
        self.inputs["from"] = {
            "button": tw_start_from_combo,
            "variable": tw_start_from_var,
        }

        tw_metric_var = tk.StringVar()
        ttk.Label(tw_widget_frame, text="Image metric:").grid(row=5, column=1)
        tw_metric_combo = ttk.Combobox(
            tw_widget_frame, textvariable=tw_metric_var, width=20
        )
        tw_metric_combo["values"] = (
            "Pixel Max",
            "Pixel Average",
            "DCT Shannon Entropy",
        )
        tw_metric_combo.state(["readonly"])
        tw_metric_combo.grid(row=6, column=1, pady=5)
        tw_metric_combo.current(0)
        self.inputs["metric"] = {"button": tw_metric_combo, "variable": tw_metric_var}

        tw_fitfunc_var = tk.StringVar()
        ttk.Label(tw_widget_frame, text="Fit func:").grid(row=7, column=1)
        tw_fitfunc_combo = ttk.Combobox(
            tw_widget_frame, textvariable=tw_fitfunc_var, width=20
        )
        tw_fitfunc_combo["values"] = (
            "poly",
            "gauss",
        )
        tw_fitfunc_combo.state(["readonly"])
        tw_fitfunc_combo.grid(row=8, column=1, pady=5)
        tw_fitfunc_combo.current(0)
        self.inputs["fitfunc"] = {
            "button": tw_fitfunc_combo,
            "variable": tw_fitfunc_var,
        }

        #: ttk.Button: Tony Wilson Button
        self.tony_wilson_button = ttk.Button(tw_widget_frame, text="RUN", width=15)
        self.tony_wilson_button.grid(row=9, column=1, pady=5)

        control_frame = ttk.Frame(content_frame)
        control_frame.grid(row=0, column=0)

        # Buttons
        button_frame = ttk.Frame(control_frame)
        button_frame.grid(row=0, column=0)

        #: ttk.Button: Set Button
        self.set_button = ttk.Button(button_frame, text="Set", width=15)
        self.set_button.grid(row=0, column=0, pady=5)

        #: ttk.Button: Flat Button
        self.flat_button = ttk.Button(button_frame, text="Flat", width=15)
        self.flat_button.grid(row=1, column=0, pady=5)

        #: ttk.Button: Zero Button
        self.zero_button = ttk.Button(button_frame, text="Zero", width=15)
        self.zero_button.grid(row=2, column=0, pady=5)

        #: ttk.Button: Clear Button
        self.clear_button = ttk.Button(button_frame, text="Clear All", width=15)
        self.clear_button.grid(row=3, column=0, pady=5)

        #: ttk.Button: Save Button
        self.save_wcs_button = ttk.Button(button_frame, text="Save WCS File", width=15)
        self.save_wcs_button.grid(row=0, column=1, pady=5)

        #: ttk.Button: Load Button
        self.from_wcs_button = ttk.Button(button_frame, text="From WCS File", width=15)
        self.from_wcs_button.grid(row=1, column=1, pady=5)

        #: ttk.Button: Select All Button
        self.select_all_modes = ttk.Button(button_frame, text="Select All", width=15)
        self.select_all_modes.grid(row=2, column=1, pady=5)

        #: ttk.Button: Deselect All Button
        self.deselect_all_modes = ttk.Button(
            button_frame, text="Deselect All", width=15
        )
        self.deselect_all_modes.grid(row=3, column=1, pady=5)

        scroll = ScrollFrame(control_frame)

        for i in range(self.n_modes):
            mode_name = self.mode_names[i]

            self.mode_labels[mode_name] = ttk.Label(
                scroll.frame, text=self.mode_names[i]
            )
            self.mode_labels[mode_name].grid(row=i, column=0)

            mode_check_var = tk.BooleanVar()
            mode_check_var.set(False)
            mode_check = ttk.Checkbutton(scroll.frame, variable=mode_check_var)
            mode_check.grid(row=i, column=1)
            self.modes_armed[mode_name] = {
                "button": mode_check,
                "variable": mode_check_var,
            }

            self.inputs[mode_name] = LabelInput(scroll.frame, input_args={"width": 10})
            self.inputs[mode_name].set(0.0)
            self.inputs[mode_name].grid(row=i, column=2)

        scroll.grid(row=1, column=0)

        save_frame = ttk.Frame(control_frame)
        save_frame.grid(row=2, column=0)

        # ttk.Checkbutton: save detailed report at the end
        save_report_var = tk.BooleanVar()
        save_report_check = ttk.Checkbutton(save_frame, variable=save_report_var)
        save_report_check.grid(row=0, column=0)
        self.inputs["save_report"] = {
            "button": save_report_check,
            "variable": save_report_var,
        }
        ttk.Label(save_frame, text="Save detailed report?").grid(row=0, column=1)

        #: ttk.Frame: Plot Frame
        self.plot_frame = ttk.Frame(master=content_frame)
        self.plot_frame.grid(row=0, column=1, rowspan=2)

        #: matplotlib.figure.Figure: Figure
        self.fig = Figure(figsize=(3, 5), dpi=100)

        #: matplotlib.axes.Axes: Mirror Image
        self.mirror_img = self.fig.add_subplot(211)

        #: matplotlib.axes.Axes: Coefficients Bar
        self.coefs_bar = self.fig.add_subplot(212)
        self.fig.tight_layout()

        canvas = FigureCanvasTkAgg(self.fig, master=self.plot_frame)
        canvas.draw()
        canvas.get_tk_widget().grid(
            row=0, column=0, sticky=(tk.NSEW), padx=(5, 5), pady=(5, 5)
        )

        #: matplotlib.figure.Figure: Figure
        self.fig_tw = Figure(figsize=(4, 5), dpi=100)

        #: matplotlib.axes.Axes: Peaks Plot
        self.peaks_plot = self.fig_tw.add_subplot(211)

        #: matplotlib.axes.Axes: Trace Plot
        self.trace_plot = self.fig_tw.add_subplot(212)
        self.fig_tw.tight_layout()

        canvas = FigureCanvasTkAgg(self.fig_tw, master=self.tab_tw)
        canvas.draw()
        canvas.get_tk_widget().grid(
            row=0, column=0, sticky=(tk.NSEW), padx=(5, 5), pady=(5, 5)
        )

        camera_var = tk.StringVar()
        #: ttk.Combobox: Camera List
        self.camera_list = ttk.Combobox(master=self.tab_cnn, textvariable=camera_var)
        # self.camera_list["values"] = ("cam_0", "cam_1")
        self.camera_list.grid(row=0, column=0, padx=10, pady=10)
        # self.camera_list.current(0)

    def onFrameConfigure(self, event):
        """Reset the scroll region to encompass the inner frame.

        Parameters
        ----------
        event : tk.Event
            Event
        """
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))

    def get_widgets(self):
        """Get widgets

        Returns
        -------
        dict
            Dictionary of widgets
        """
        return self.inputs

    def get_labels(self):
        """Get labels

        Returns
        -------
        dict
            Dictionary of labels
        """
        return self.mode_labels

    def get_modes_armed(self):
        """Get armed modes

        Returns
        -------
        dict
            Dictionary of armed modes
        """
        return self.modes_armed

    def set_widgets(self, coefs):
        """Set widgets

        Parameters
        ----------
        coefs : list
            List of coefficients
        """
        for i, c in enumerate(coefs):
            self.inputs[self.mode_names[i]].set(c)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
import tkinter.ttk as ttk
import platform

# Third Party Imports

# Local Imports


class ScrolledFrame(ttk.Frame):
    """A scrollable frame implemented in tkinter.

    Based upon: https://web.archive.org/web/20170514022131id_/http://tkinter.unpythonic
    .net/wiki/VerticalScrolledFrame
    """

    def __init__(self, parent, *args, **kw):
        """Initialize the ScrolledFrame.

        Parameters
        ----------
        root : Tk top-level widget
            Tkinter GUI instance to which this ScrolledFrame belongs.
        *args
            Additional arguments to pass to the ttk.Frame constructor.
        **kw
            Additional keyword arguments to pass to the ttk.Frame constructor.
        """
        ttk.Frame.__init__(self, parent, *args, **kw)

        # Create a canvas object and a vertical scrollbar for scrolling it.
        vscrollbar = ttk.Scrollbar(self, orient=tk.constants.VERTICAL)
        vscrollbar.pack(
            fill=tk.constants.Y, side=tk.constants.RIGHT, expand=tk.constants.FALSE
        )
        hscrollbar = ttk.Scrollbar(self, orient=tk.constants.HORIZONTAL)
        hscrollbar.pack(
            fill=tk.constants.X, side=tk.constants.BOTTOM, expand=tk.constants.FALSE
        )
        #: tk.Canvas: The canvas object for the ScrolledFrame.
        self.canvas = tk.Canvas(
            self,
            bd=0,
            highlightthickness=0,
            yscrollcommand=vscrollbar.set,
            xscrollcommand=hscrollbar.set,
            scrollregion=(0, 0, 100, 100),
        )
        self.canvas.pack(
            side=tk.constants.LEFT, fill=tk.constants.BOTH, expand=tk.constants.TRUE
        )
        vscrollbar.config(command=self.canvas.yview)
        hscrollbar.config(command=self.canvas.xview)

        # Reset the view
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)

        # Create a frame inside the canvas which will be scrolled with it.
        #: ttk.Frame: The interior frame of the ScrolledFrame.
        self.interior = interior = ttk.Frame(self.canvas)
        _ = self.canvas.create_window(0, 0, window=interior, anchor=tk.constants.NW)

        # Track changes to the canvas and frame width and sync them,
        # also updating the scrollbar.
        def _configure_interior(event):
            """Configure the interior frame based on size changes.

            Updates the scrollbars to match the size of the inner frame, and updates
            the canvas width to fit the inner frame.

            Parameters
            ----------
            event : Tkinter Event
                The event triggering the configuration.
            """
            size = (interior.winfo_reqwidth(), interior.winfo_reqheight())
            self.canvas.config(scrollregion="0 0 %s %s" % size)
            if interior.winfo_reqwidth() != self.canvas.winfo_width():
                self.canvas.config(width=interior.winfo_reqwidth())
            if interior.winfo_reqheight() != self.canvas.winfo_reqheight():
                self.canvas.config(height=interior.winfo_reqheight())

        interior.bind("<Configure>", _configure_interior)

    def mouse_wheel(self, event):
        """Handle the mouse wheel event for scrolling.

        Parameters
        ----------
        event : Tkinter Event
            The mouse wheel event.
        """
        delta = 120 if platform.system() != "Darwin" else 1
        shift = -1 * int(event.delta / delta)
        # TODO: event.state may only work on Mac (Darwin). Investigate for Windows.
        if event.state == 0:
            self.canvas.yview_scroll(shift, "units")
        elif event.state == 1:
            self.canvas.xview_scroll(shift, "units")



import threading
import time

from pyvcam import pvc
from pyvcam.camera import Camera

NUM_FRAMES = 20

class TriggerThreadRun (threading.Thread):
    def __init__(self, cam):
        threading.Thread.__init__(self)
        self.cam = cam
        self.end = False

    def run(self):
        while not self.end:
            try:
                self.cam.sw_trigger()
                print('SW Trigger success')
                time.sleep(0.05)
            except Exception as e:
                pass

    def stop(self):
        self.end = True

def collectFrames(cam):
    framesReceived = 0

    while framesReceived < NUM_FRAMES:
        time.sleep(0.005)

        try:
            frame, fps, frame_count = cam.poll_frame()
            print('Count: {} FPS: {} First five pixels of frame: {}'.format(frame_count, round(fps, 2), frame['pixel_data'][0, 0:5]))
            framesReceived += 1
        except Exception as e:
            print(str(e))

    return framesReceived

def main():
    # Initialize PVCAM and find the first available camera.
    pvc.init_pvcam()

    cam = [cam for cam in Camera.detect_camera()][0]
    cam.open()
    cam.speed_table_index = 0
    cam.exp_mode = 'Software Trigger Edge'

    # Start a thread for executing the trigger
    t1 = TriggerThreadRun(cam)
    t1.start()

    # Collect frames in live mode
    cam.start_live()
    framesReceived = collectFrames(cam)
    cam.finish()
    print('Received live frames: ' + str(framesReceived) + '\n')

    # Collect frames in sequence mode
    cam.start_seq(num_frames=NUM_FRAMES)
    framesReceived = collectFrames(cam)
    cam.finish()
    print('Received seq frames: ' + str(framesReceived) + '\n')

    t1.stop()
    cam.close()
    pvc.uninit_pvcam()

    print('Done')

if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
import tkinter as tk
from tkinter import ttk

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedSpinbox, ValidatedEntry

# Logging Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class TilingWizardPopup:
    """Popup for tiling parameters in View."""

    def __init__(self, root, *args, **kwargs):
        """Initialize the popup

        Parameters
        ----------
        root : tk.Tk
            The root Tk instance
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.
        """
        #: tk.TopLevel: The popup window
        self.popup = PopUp(
            root,
            "Multi-Position Tiling Wizard",
            "800x250+10+10",
            top=False,
            transient=False,
        )

        # Grab the axes we want to grid out. Optionally,
        # we can pass these in the kwargs
        axes = list(kwargs.get("axes", "XYZF"))

        nrow = len(axes) + 2  # +2 for percent overlay, total number of tiles, populate
        # <set x-start> [x-start] <set-x-end> [x-end]
        # [x-distance] [x-fov-width] [x-tiles]
        ncol = 7

        # Set up the grid on the popup's frame
        for col in range(ncol):
            self.popup.content_frame.columnconfigure(col, pad=5, weight=1)
        for row in range(nrow):
            self.popup.content_frame.rowconfigure(row, pad=5, weight=1)
        #: dict: The GUI elements used for input, mostly ttk.Frames
        self.inputs = {}
        #: dict: The GUI elements used for buttons, mostly ttk.Buttons
        self.buttons = {}

        # Add one row per axis
        for row, ax in enumerate(axes):
            # Set start position
            start_var = f"{ax.lower()}_start"
            self.buttons[start_var] = ttk.Button(
                self.popup.content_frame, text=f"Set {ax.upper()} Start"
            )
            self.buttons[start_var].grid(
                row=row, column=0, sticky=tk.NSEW, padx=(5, 0), pady=(5, 0)
            )
            self.inputs[start_var] = LabelInput(
                parent=self.popup.content_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
                input_args={"width": 5},
            )
            self.inputs[start_var].grid(
                row=row, column=1, sticky=tk.NSEW, pady=(5, 0), padx=(5, 0)
            )
            self.inputs[start_var].widget.state(["disabled"])

            # Set end position
            end_var = f"{ax.lower()}_end"
            self.buttons[end_var] = ttk.Button(
                self.popup.content_frame, text=f"Set {ax.upper()} End"
            )
            self.buttons[end_var].grid(
                row=row, column=2, sticky=tk.NSEW, padx=(5, 0), pady=(5, 0)
            )
            self.inputs[end_var] = LabelInput(
                parent=self.popup.content_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
                input_args={"width": 5},
            )
            self.inputs[end_var].grid(
                row=row, column=3, sticky=tk.NSEW, pady=(5, 0), padx=(5, 0)
            )
            self.inputs[end_var].widget.state(["disabled"])

            # Distance between start and end of this axis
            dist_var = f"{ax.lower()}_dist"
            self.inputs[dist_var] = LabelInput(
                parent=self.popup.content_frame,
                label=f"{ax.upper()} Distance",
                input_class=ValidatedEntry,
                input_var=tk.StringVar(),
                input_args={"width": 5},
            )
            self.inputs[dist_var].grid(
                row=row, column=4, sticky=tk.NSEW, pady=(5, 0), padx=(5, 0)
            )
            self.inputs[dist_var].widget.state(["disabled"])

            # FOV width for this particular axis (e.g. for X it is the number
            # of vertical pixels on the camera multipled by the effective
            # pixel size)
            fov_var = f"{ax.lower()}_fov"
            self.inputs[fov_var] = LabelInput(
                parent=self.popup.content_frame,
                label=f"{ax.upper()} FOV Dist",
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
                input_args={
                    "width": 5,
                    "from_": "-Infinity",
                    "to": "Infinity",
                    "increment": 100.0,
                },
            )
            self.inputs[fov_var].grid(
                row=row, column=5, sticky=tk.NSEW, pady=(5, 0), padx=(5, 0)
            )
            # self.inputs[fov_var].widget.state(["disabled"])

            # Number of tiles, roughly ceil((ax_end - ax_start) / fov_width)
            # (not accounting for percent overlap)
            tiles_var = f"{ax.lower()}_tiles"
            self.inputs[tiles_var] = LabelInput(
                parent=self.popup.content_frame,
                label="Num. Tiles",
                input_class=ValidatedEntry,
                input_var=tk.StringVar(),
                input_args={"width": 5},
            )
            self.inputs[tiles_var].grid(
                row=row, column=6, sticky=tk.NSEW, pady=(5, 0), padx=(5, 0)
            )
            self.inputs[tiles_var].widget.state(["disabled"])

        # In the final row, add percent overlap, total number of tiles, populate button
        self.inputs["percent_overlap"] = LabelInput(
            parent=self.popup.content_frame,
            label="% Overlap",
            input_class=ValidatedSpinbox,
            input_var=tk.StringVar(),
            input_args={"width": 5, "increment": 5, "from_": 0, "to": 100},
        )
        self.inputs["percent_overlap"].grid(
            row=len(axes), column=5, sticky=tk.NSEW, padx=(5, 0), pady=(5, 0)
        )

        self.inputs["total_tiles"] = LabelInput(
            parent=self.popup.content_frame,
            label="Total Tiles",
            input_class=ValidatedEntry,
            input_var=tk.StringVar(),
            input_args={"width": 5},
        )
        self.inputs["total_tiles"].widget.state(["disabled"])
        self.inputs["total_tiles"].grid(
            row=len(axes), column=6, sticky=tk.NSEW, padx=(5, 0), pady=(5, 0)
        )

        self.buttons["set_table"] = ttk.Button(
            self.popup.content_frame, text="Populate Multi-Position Table"
        )
        self.buttons["set_table"].grid(
            row=len(axes) + 1,
            column=5,
            columnspan=2,
            sticky=tk.NE,
            padx=(0, 5),
            pady=(5, 0),
        )

    # Getters
    def get_variables(self):
        """Get the variables tied to the widgets

        This function returns a dictionary of all the variables that are tied to each
        widget name. The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            A dictionary of all the variables that are tied to each widget name.
        """
        return {key: widget.get_variable() for key, widget in self.inputs.items()}

    def get_widgets(self):
        """Get the widgets

        This function returns the dictionary that holds the input widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        self.inputs : dict
            A dictionary of all the widgets that are tied to each widget name.
        """
        return self.inputs

    def get_buttons(self):
        """Get the buttons

        This function returns the dictionary that holds the buttons.
        The key is the button name, value is the button.

        Returns
        -------
        self.buttons : dict
            A dictionary of all the buttons that are tied to each button name.
        """
        return self.buttons



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
from tkinter import NSEW, NW, ttk
import tkinter
import logging

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.popup import PopUp

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ilastik_setting_popup:
    """Popup window for setting up ilastik segmentation settings."""

    def __init__(self, root, *args, **kwargs):
        """Initialize the popup window

        Parameters
        ----------
        root : tkinter.Tk
            Root window
        *args :
            Variable length argument list
        **kwargs :
            Arbitrary keyword arguments
        """

        # Creating popup window with this name and size/placement, PopUp is a
        #: PopUp: PopUp window class
        self.popup = PopUp(
            root, "Ilastik Settings", "+320+180", top=False, transient=True
        )

        # Creating content frame
        content_frame = self.popup.get_frame()

        #: tkinter.StringVar: Variable for project name
        self.project_name_var = tkinter.StringVar(
            value="Please select an ilastik pixelclassification project file!"
        )

        project_name_entry = ttk.Entry(
            content_frame,
            textvariable=self.project_name_var,
            width=50,
            state="readonly",
        )
        project_name_entry.grid(row=1, column=0, pady=(10, 10), padx=(0, 10))
        #: ttk.Button: Button for loading ilastik project
        self.load_project_btn = ttk.Button(
            content_frame, text="Load Ilastik Project", width=30
        )
        self.load_project_btn.grid(row=1, column=1, pady=(10, 10), padx=(0, 10))

        tkinter.Label(content_frame, text="Select all target labels:").grid(
            row=2, sticky=NW
        )
        #: ttk.Frame: Frame for labels
        self.label_frame = ttk.Frame(content_frame, padding="3 3 12 12")
        self.label_frame.grid(row=3, columnspan=2, sticky=NSEW)
        default_colors = ["red", "#0082c8", "#ffe119"]
        for i in range(3):
            label_widget = ttk.Checkbutton(self.label_frame, text="Label" + str(i + 1))
            label_widget.grid(
                row=1 + i, column=0, pady=(0, 10), padx=(20, 5), sticky="W"
            )
            color_block = tkinter.Label(
                self.label_frame, background=default_colors[i], width=3, height=1
            )
            color_block.grid(row=1 + i, column=1, pady=(0, 10), padx=(0, 10))

        # segmentation usage
        tkinter.Label(content_frame, text="Choose the way to use segmentation:").grid(
            row=4, sticky=NW
        )
        #: tkinter.Checkbutton: Checkbutton for showing segmentation
        self.show_on_gui = tkinter.Checkbutton(content_frame, text="Show Segmentation")
        self.show_on_gui.grid(row=5, column=0, padx=(20, 0), pady=(10, 20), sticky="W")

        #: tkinter.Checkbutton: Checkbutton for marking position
        self.mark_position = tkinter.Checkbutton(content_frame, text="Mark Position")
        self.mark_position.grid(row=5, column=1, padx=(0, 0), pady=(10, 20), sticky="W")

        #: ttk.Button: Button for confirming setting
        self.confirm_btn = ttk.Button(content_frame, text="Confirm Setting", width=30)
        self.confirm_btn.grid(row=7, column=1, sticky="SE", padx=(0, 10), pady=(0, 10))

    def get_buttons(self):
        """Get all buttons

        Returns
        -------
        dict
            Dictionary of all buttons
        """
        return {"load": self.load_project_btn, "confirm": self.confirm_btn}

    def get_variables(self):
        """Get all variables

        Returns
        -------
        dict
            Dictionary of all variables
        """
        return {"project_name": self.project_name_var}

    def get_widgets(self):
        """Get all widgets

        Returns
        -------
        dict
            Dictionary of all widgets
        """
        return {
            "label_frame": self.label_frame,
            "show_segmentation": self.show_on_gui,
            "mark_position": self.mark_position,
        }



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import argparse
from pathlib import Path

# Third Party Imports

# Local Imports
from navigate.config import get_configuration_paths


def evaluate_parser_input_arguments(args):
    """If non-default configuration, experiment, or ETL constant file is provided as an
    input argument.

    Accepts a ArgumentParser object
    Retrieves the default configuration/experiment/etc paths.
    Adds  non-default parser arguments to the ArgumentParser object
    Parses the object, and returns the desired configuration/experiment/etc paths.

    Parameters
    ----------
    args : argparse.Namespace
        Dictionary of parser input arguments

    Returns
    -------
    configuration_path : str
        Path to configuration file.
    experiment_path : str
        Path to experiment file
    waveform_constants_path,
        Path to remote focusing and galvo waveform constants file
    rest_api_path
        Path to REST API file
    waveform_templates_path
        Path to waveform templates file
    logging_path
        Path to non-default logging location
    """
    # Retrieve the Default Configuration paths
    (
        configuration_path,
        experiment_path,
        waveform_constants_path,
        rest_api_path,
        waveform_templates_path,
    ) = get_configuration_paths()

    # Evaluate Input Arguments
    if args.config_file:
        assert args.config_file.exists(), "Configuration file Path {} not valid".format(
            args.config_file
        )
        configuration_path = args.config_file

    if args.experiment_file:
        assert (
            args.experiment_file.exists()
        ), "experiment_file file Path {} not valid".format(args.experiment_file)
        experiment_path = args.experiment_file

    if args.waveform_constants_path:
        assert (
            args.waveform_constants_path.exists()
        ), "waveform_constants_path Path {} not valid".format(
            args.waveform_constants_path
        )
        waveform_constants_path = args.waveform_constants_path

    if args.rest_api_file:
        assert args.rest_api_file.exists(), "rest_api_file Path {} not valid".format(
            args.rest_api_file
        )
        rest_api_path = args.rest_api_file

    if args.waveform_templates_file:
        assert (
            args.waveform_templates_file.exists()
        ), "waveform_templates Path {} not valid".format(args.waveform_templates_file)
        waveform_templates_path = args.waveform_templates_file

    # Creating Loggers etc., they exist globally so no need to pass
    if args.logging_config:
        assert args.logging_config.exists(), "Logging Config Path {} not valid".format(
            args.logging_config
        )
        logging_path = args.logging_config
    else:
        logging_path = None

    return (
        configuration_path,
        experiment_path,
        waveform_constants_path,
        rest_api_path,
        waveform_templates_path,
        logging_path,
    )


def create_parser():
    """Add Parser Input Arguments to ArgumentParser Object.

    Parameters
    ----------
    None

    Returns
    -------
    parser : object
        ArgumentParserObject with Added Input Arguments"""

    parser = argparse.ArgumentParser(description="navigate Command Line Arguments")

    input_args = parser.add_argument_group("Input Arguments")

    input_args.add_argument(
        "-sh",
        "--synthetic-hardware",
        required=False,
        default=False,
        action="store_true",
        help="Synthetic Hardware - "
        "Allows launching of the software without the physical devices attached.",
    )

    input_args.add_argument(
        "-d",
        required=False,
        default=False,
        action="store_true",
        help="Enables debugging tool menu to be accessible.",
    )

    # Non-Default Configuration and Experiment Input Arguments
    input_args.add_argument(
        "--config-file",
        type=Path,
        required=False,
        default=None,
        help="Non-default path to the configuration.yml file.  \n"
        "This file sets the default global physical"
        "state of the microscope.  For example, number of lasers, number of cameras, "
        "digital I/O, analog I/O, etc.",
    )

    input_args.add_argument(
        "--experiment-file",
        type=Path,
        required=False,
        default=None,
        help="Non-default path to the experiment.yml file. \n"
        "This file sets the default user-specified "
        "physical state of the microscope. For example, how many channels, "
        "and what are their exposure time, filter position, laser, etc. ",
    )

    input_args.add_argument(
        "--waveform-constants-path",
        type=Path,
        required=False,
        default=None,
        help="Non-default path to the waveform_constants.yml file.  \n"
        "This file specifies the wavelength- and zoom-specific amplitude and offset "
        "of the ETL waveform generation.",
    )

    input_args.add_argument(
        "--rest-api-file",
        type=Path,
        required=False,
        default=None,
        help="Non-default path to the rest_api_config.yml config file \n"
        "This file specifies urls of restful api services.",
    )

    input_args.add_argument(
        "--waveform-templates-file",
        type=Path,
        required=False,
        default=None,
        help="Non-default path to the waveform_templates.yml config file \n"
        "This file specifies all waveform templates.",
    )

    input_args.add_argument(
        "--logging-config",
        type=Path,
        required=False,
        default=None,
        help="Non-default path to the logging.yml config file \n"
        "This file specifies how the logging will be performed.",
    )

    return parser



"""
API for connection to Thorlabs.MotionControl.KCube.InertialMotor.dll.
See Thorlabs.MotionControl.KCube.InertialMotor.h for more functions to implement.

Tested on Thorlabs KIM001.
"""


import ctypes
import ctypes.wintypes
from enum import IntEnum

CODING = "ascii"

__dll = ctypes.WinDLL(
    "C:\Program Files\Thorlabs\Kinesis\Thorlabs.MotionControl.KCube.InertialMotor.dll"
)


class TLFTDICommunicationError(Exception):
    """Exception for Thorlabs FTDI communications module or supporting code."""


class TLDLLError(Exception):
    """Exception for general Thorlabs DLL control errors."""


class TLMotorDLLError(Exception):
    """Exception for motor-specific Thorlabs DLL errors."""


class FT_Status(IntEnum):
    """The following errors are generated from the FTDI communications module,
    supporting code or device libraries."""

    FT_OK = 0  # No error
    FT_InvalidHandle = 1  # The FTDI functions have not been initialized.
    FT_DeviceNotFound = 2  # The Device could not be found. This can be generated if the
    # function TLI_BuildDeviceList() has not been called.
    FT_DeviceNotOpened = 3  # The Device must be opened before it can be accessed
    FT_IOError = 4  # An I/O Error has occured in the FTDI chip.
    FT_InsufficientResources = (
        5  # There are Insufficient resources to run this application.
    )
    FT_InvalidParameter = 6  # An invalid parameter has been supplied to the device.
    FT_DeviceNotPresent = 7  # The Device is no longer present. The device may have been
    # disconnected since the last TLI_BuildDeviceList() call.
    FT_IncorrectDevice = 8  # The device detected does not match that expected
    FT_NoDLLLoaded = 16  # The library for this device could not be found
    FT_NoFunctionsAvailable = 17  # No functions available for this device
    FT_FunctionNotAvailable = 18  # The function is not available for this device.
    FT_BadFunctionPointer = 19  # Bad function pointer detected
    FT_GenericFunctionFail = 20  # The function failed to complete successfully.
    FT_SpecificFunctionFail = 21  # The function failed to complete successfully.


FT_Status_Description = {
    FT_Status.FT_OK: "No error",
    FT_Status.FT_InvalidHandle: "The FTDI functions have not been initialized.",
    FT_Status.FT_DeviceNotFound: "The Device could not be found. This can be generated "
    "if the function TLI_BuildDeviceList() has not been called.",
    FT_Status.FT_DeviceNotOpened: "The Device must be opened before it can "
    "be accessed.",
    FT_Status.FT_IOError: "An I/O Error has occured in the FTDI chip.",
    FT_Status.FT_InsufficientResources: "There are Insufficient resources to run "
    "this application.",
    FT_Status.FT_InvalidParameter: "An invalid parameter has been supplied to "
    "the device.",
    FT_Status.FT_DeviceNotPresent: "The Device is no longer present. The device may "
    "have been disconnected "
    "since the last TLI_BuildDeviceList() call.",
    FT_Status.FT_IncorrectDevice: "The device detected does not match that expected.",
    FT_Status.FT_NoDLLLoaded: "The library for this device could not be found.",
    FT_Status.FT_FunctionNotAvailable: "The function is not available for this device.",
    FT_Status.FT_NoFunctionsAvailable: "No functions available for this device.",
    FT_Status.FT_BadFunctionPointer: "Bad function pointer detected.",
    FT_Status.FT_GenericFunctionFail: "The function failed to complete successfully.",
    FT_Status.FT_SpecificFunctionFail: "The function failed to complete successfully.",
}


class TL_DLL_Error(IntEnum):
    """The following errors are general errors generated by all DLLs."""

    TL_ALREADY_OPEN = 32  # Attempt to open a device that was already open.
    TL_NO_RESPONSE = 33  # The device has stopped responding.
    TL_NOT_IMPLEMENTED = 34  # This function has not been implemented.
    TL_FAULT_REPORTED = 35  # The device has reported a fault.
    TL_INVALID_OPERATION = 36  # The function could not be completed at this time.
    TL_DISCONNECTING = (
        40  # The function could not be completed because the device is disconnected.
    )
    TL_FIRMWARE_BUG = 41  # The firmware has thrown an error.
    TL_INITIALIZATION_FAILURE = 42  # The device has failed to initialize
    TL_INVALID_CHANNEL = 43  # An Invalid channel address was supplied


TL_DLL_Error_Description = {
    TL_DLL_Error.TL_ALREADY_OPEN: "Attempt to open a device that was already open.",
    TL_DLL_Error.TL_NO_RESPONSE: "The device has stopped responding.",
    TL_DLL_Error.TL_NOT_IMPLEMENTED: "This function has not been implemented.",
    TL_DLL_Error.TL_FAULT_REPORTED: "The device has reported a fault.",
    TL_DLL_Error.TL_INVALID_OPERATION: "The function could not be completed at "
    "this time.",
    TL_DLL_Error.TL_DISCONNECTING: "The function could not be completed because the "
    "device is disconnected.",
    TL_DLL_Error.TL_FIRMWARE_BUG: "The firmware has thrown an error.",
    TL_DLL_Error.TL_INITIALIZATION_FAILURE: "The device has failed to initialize.",
    TL_DLL_Error.TL_INVALID_CHANNEL: "An Invalid channel address was supplied.",
}


class Motor_DLL_Error(IntEnum):
    """The following errors are motor specific errors generated by the Motor DLLs."""

    TL_UNHOMED = 37  # The device cannot perform this function until it has been Homed.
    TL_INVALID_POSITION = 38  # The function cannot be performed as it would result in
    # an illegal position.
    TL_INVALID_VELOCITY_PARAMETER = 39  # An invalid velocity parameter was supplied.
    # The velocity must be greater than zero.
    TL_CANNOT_HOME_DEVICE = 44  # This device does not support Homing. Check the Limit
    # switch parameters are correct.
    TL_JOG_CONTINOUS_MODE = 45  # An invalid jog mode was supplied for the jog function.
    TL_NO_MOTOR_INFO = (
        46  # There is no Motor Parameters available to convert Real World Units.
    )
    TL_CMD_TEMP_UNAVAILABLE = 47  # Command temporarily unavailable, Device may be busy.


Motor_DLL_Error_Description = {
    Motor_DLL_Error.TL_UNHOMED: "The device cannot perform this function until it has "
    "been Homed.",
    Motor_DLL_Error.TL_INVALID_POSITION: "The function cannot be performed as it would "
    "result in an illegal position.",
    Motor_DLL_Error.TL_INVALID_VELOCITY_PARAMETER: "An invalid velocity parameter was "
    "supplied. The velocity must be greater than zero.",
    Motor_DLL_Error.TL_CANNOT_HOME_DEVICE: "This device does not support Homing. Check "
    "the Limit switch parameters are correct.",
    Motor_DLL_Error.TL_JOG_CONTINOUS_MODE: "An invalid jog mode was supplied for the "
    "jog function.",
    Motor_DLL_Error.TL_NO_MOTOR_INFO: "There is no Motor Parameters available to "
    " convert Real World Units.",
    Motor_DLL_Error.TL_CMD_TEMP_UNAVAILABLE: "Command temporarily unavailable, Device "
    "may be busy.",
}


def in_enum(value, enum):
    values = set(item.value for item in enum)
    return value in values


def errcheck(result, func, args):
    """
    Wraps the call to DLL functions.

    Parameters
    ----------
    result : ctypes.c_short
        Error code or 0 if successful.
    func : function
        DLL function
    args : tuple
        Arguments passed to the DLL function, defined in argtypes

    Returns
    -------
    result : int
        Error code or 0 if successful.
    """
    if result:
        # returned a non-zero value
        if in_enum(result, FT_Status):
            raise TLFTDICommunicationError(FT_Status_Description[result])
        elif in_enum(result, TL_DLL_Error):
            raise TLDLLError(TL_DLL_Error_Description[result])
        elif in_enum(result, Motor_DLL_Error):
            raise TLMotorDLLError(Motor_DLL_Error_Description[result])
        else:
            raise Exception(f"Unknown error {result} in Thorlabs device.")
    return int(result)


__dll.TLI_BuildDeviceList.restype = ctypes.c_short
__dll.TLI_BuildDeviceList.errcheck = errcheck


def TLI_BuildDeviceList():
    """This function builds an internal collection of all devices found on the USB that
    are not currently open.

    Returns
    -------
    int
        The error code or 0 if successful.
    """
    return __dll.TLI_BuildDeviceList()


__dll.TLI_GetDeviceListSize.restype = ctypes.c_short


def TLI_GetDeviceListSize():
    """
    Gets the device list size.

    Returns
    -------
    int
        Number of devices in device list.
    """
    return __dll.TLI_GetDeviceListSize()


__dll.TLI_GetDeviceListExt.argtypes = [ctypes.c_char_p, ctypes.wintypes.DWORD]
__dll.TLI_GetDeviceListExt.restype = ctypes.c_short
__dll.TLI_GetDeviceListExt.errcheck = errcheck


def TLI_GetDeviceListExt():
    """Get the entire contents of the device list.

    Returns
    -------
    List
        List of device serial numbers as strings.
    """

    c_buf_len = 256
    c_buf = ctypes.create_string_buffer(c_buf_len)
    __dll.TLI_GetDeviceListExt(c_buf, c_buf_len)

    return str(c_buf.value.decode(CODING)).split(",")[:-1]


__dll.TLI_GetDeviceListByTypeExt.argtypes = [
    ctypes.c_char_p,
    ctypes.wintypes.DWORD,
    ctypes.c_int,
]
__dll.TLI_GetDeviceListByTypeExt.restype = ctypes.c_short
__dll.TLI_GetDeviceListByTypeExt.errcheck = errcheck


def TLI_GetDeviceListByTypeExt(type_id):
    """Get the contents of the device list which match the supplied type_id.

    Parameters
    ----------
    type_id : int
        Device type.

    Returns
    -------
    List
        List of device serial numbers as strings.
    """

    c_buf_len = 256
    c_buf = ctypes.create_string_buffer(c_buf_len)
    __dll.TLI_GetDeviceListByTypeExt(c_buf, c_buf_len, type_id)

    return str(c_buf.value.decode(CODING)).split(",")[:-1]


class MOT_MotorTypes(IntEnum):
    """Values that represent THORLABSDEVICE_API."""

    MOT_NotMotor = 0
    MOT_DCMotor = 1
    MOT_StepperMotor = 2
    MOT_BrushlessMotor = 3
    MOT_CustomMotor = 100


class TLI_DeviceInfo(ctypes.Structure):
    """Information about the device generated from serial number."""

    _pack_ = 1
    _fields_ = [
        ("typeID", ctypes.wintypes.DWORD),  # The device Type ID
        ("description", ctypes.c_char * 65),  # The device description.
        ("serialNo", ctypes.c_char * 16),  # The device serial number.
        ("PID", ctypes.wintypes.DWORD),  # The USB PID number.
        (
            "isKnownType",
            ctypes.wintypes.BOOL,
        ),  # True if this object is a type known to the Motion Control software.
        ("motorType", ctypes.c_int),  # The motor type (if a motor)
        (
            "isPiezoDevice",
            ctypes.wintypes.BOOL,
        ),  # True if the device is a piezo device.
        ("isLaser", ctypes.wintypes.BOOL),  # True if the device is a laser.
        ("isCustomType", ctypes.wintypes.BOOL),  # True if the device is a custom type.
        ("isRack", ctypes.wintypes.BOOL),  # True if the device is a rack.
        (
            "maxChannels",
            ctypes.c_short,
        ),  # Defines the number of channels available in this device.
    ]


__dll.TLI_GetDeviceInfo.argtypes = [ctypes.c_char_p, ctypes.POINTER(TLI_DeviceInfo)]
__dll.TLI_GetDeviceInfo.restype = ctypes.c_short


def TLI_GetDeviceInfo(serial_no):
    """Get the device information from the USB port.

    The Device Info is read from the USB port not from the device itself.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.

    Returns
    -------
    TLI_DeviceInfo
        The device information.
    """

    info = TLI_DeviceInfo()

    __dll.TLI_GetDeviceInfo(
        serial_no.encode(CODING), ctypes.byref(info)
    )  # 1 if successful, 0 if not

    return info


__dll.KIM_Open.argtypes = [ctypes.c_char_p]
__dll.KIM_Open.restype = ctypes.c_short
__dll.KIM_Open.errcheck = errcheck


def KIM_Open(serial_no):
    """Open the device for communications.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.

    Returns
    -------
    int
        The error code or 0 if successful.
    """

    return __dll.KIM_Open(serial_no.encode(CODING))


__dll.KIM_Close.argtypes = [ctypes.c_char_p]


def KIM_Close(serial_no):
    """Disconnect and close the device.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.

    Returns
    -------
    None
    """

    __dll.KIM_Close(serial_no.encode(CODING))


__dll.KIM_StartPolling.argtypes = [ctypes.c_char_p, ctypes.c_int]
__dll.KIM_StartPolling.restype = ctypes.wintypes.BOOL


def KIM_StartPolling(serial_no, milliseconds):
    """Starts the internal polling loop which continuously requests position and status.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.
    milliseconds : int
        The milliseconds polling rate.

    Returns
    -------
    bool
        True if successful, False otherwise
    """
    return bool(__dll.KIM_StartPolling(serial_no.encode(CODING), milliseconds))


__dll.KIM_StopPolling.argtypes = [ctypes.c_char_p]


def KIM_StopPolling(serial_no):
    """Stops the internal polling loop.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.

    Returns
    -------
    None
    """
    __dll.KIM_StopPolling(serial_no.encode(CODING))


class KIM_Channels(IntEnum):  # unsigned short
    Channel1 = 1
    Channel2 = 2
    Channel3 = 3
    Channel4 = 4


__dll.KIM_RequestCurrentPosition.argtypes = [ctypes.c_char_p, ctypes.c_ushort]
__dll.KIM_RequestCurrentPosition.restype = ctypes.c_int
__dll.KIM_RequestCurrentPosition.errcheck = errcheck


def KIM_RequestCurrentPosition(serial_no, channel):
    """Gets current position.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.
    channel : int
        The device channel. One of KIM_Channels.

    Returns
    -------
    int
        The error code or 0 if successful.
    """

    return __dll.KIM_RequestCurrentPosition(serial_no.encode(CODING), channel)


__dll.KIM_GetCurrentPosition.argtypes = [ctypes.c_char_p, ctypes.c_ushort]
__dll.KIM_GetCurrentPosition.restype = ctypes.c_int


def KIM_GetCurrentPosition(serial_no, channel):
    """Gets current position.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.
    channel : int
        The device channel. One of KIM_Channels.

    Returns
    -------
    int
        Current position.
    """

    return __dll.KIM_GetCurrentPosition(serial_no.encode(CODING), channel)


__dll.KIM_MoveAbsolute.argtypes = [ctypes.c_char_p, ctypes.c_ushort, ctypes.c_int]
__dll.KIM_MoveAbsolute.restype = ctypes.c_short
__dll.KIM_MoveAbsolute.errcheck = errcheck


def KIM_MoveAbsolute(serial_no, channel, position):
    """Move absolute.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.
    channel : int
        The device channel. One of KIM_Channels.
    position : int
        The position to move to.

    Returns
    -------
    int
        The error code or 0 if successful.
    """

    return __dll.KIM_MoveAbsolute(serial_no.encode(CODING), channel, position)


__dll.KIM_SetPosition.argtypes = [ctypes.c_char_p, ctypes.c_ushort, ctypes.c_long]
__dll.KIM_SetPosition.restype = ctypes.c_short
__dll.KIM_SetPosition.errcheck = errcheck


def KIM_SetPosition(serial_no, channel, position):
    """Set the current position to position.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.
    channel : int
        The device channel. One of KIM_Channels.
    position : int
        The value the current position is mapped to.

    Returns
    -------
    int
        The error code or 0 if successful.
    """

    return __dll.KIM_SetPosition(serial_no.encode(CODING), channel, position)


class KIM_TravelDirection(IntEnum):  # byte
    Forward = 1  # An enum constant representing the forward option.
    Reverse = 2  # An enum constant representing the reverse option.


__dll.KIM_MoveJog.argtypes = [ctypes.c_char_p, ctypes.c_ushort, ctypes.c_byte]
__dll.KIM_MoveJog.restype = ctypes.c_short
__dll.KIM_MoveJog.errcheck = errcheck


def KIM_MoveJog(serial_no, channel, jog_direction):
    """Move jog.

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.
    channel : int
        The device channel. One of KIM_Channels.
    jog_direction : int
        The jog direction. One of KIM_TravelDirection.

    Returns
    -------
    int
        The error code or 0 if successful.
    """

    return __dll.KIM_MoveJog(serial_no.encode(CODING), channel, jog_direction)


__dll.KIM_MoveStop.argtypes = [ctypes.c_char_p, ctypes.c_ushort]
__dll.KIM_MoveStop.restype = ctypes.c_short
__dll.KIM_MoveAbsolute.errcheck = errcheck


def KIM_MoveStop(serial_no, channel):
    """Halt motion

    Parmeters
    ---------
    serial_number : str
        Serial number of Thorlabs Kinesis Inertial Motor (KIM) device.
    channel : int
        The device channel. One of KIM_Channels.

    Returns
    -------
    int
        The error code or 0 if successful.
    """

    return __dll.KIM_MoveStop(serial_no.encode(CODING), channel)
